import wx
from . import amulet_ui
import sys
import locale


class AmuletApp(wx.App):

    _amulet_ui: amulet_ui.AmuletUI

    def OnInit(self):
        self._amulet_ui = amulet_ui.AmuletUI(None)
        self.SetTopWindow(self._amulet_ui)
        self._amulet_ui.Show()
        return True

    def InitLocale(self):
        # https://discuss.wxpython.org/t/what-is-wxpython-doing-to-the-locale-to-makes-pandas-crash/34606/18
        if sys.version_info[:2] >= (3, 8):
            super().InitLocale()
        else:
            self.ResetLocale()
            lang, enc = locale.getlocale()
            if lang is None:
                self._initial_locale = wx.Locale(wx.LANGUAGE_DEFAULT)

    def open_level(self, path: str):
        """
        Open a level and create a tab for it.
        If a tab already exists it will just be shown.

        :param path: The path to the level to open.
        """
        self._amulet_ui.open_level(path)

    def close_level(self, path: str):
        """
        Close a level tab.

        :param path: The path to the level to close.
        """
        self._amulet_ui.close_level(path)


def get_app() -> AmuletApp:
    """Get the app instance."""
    app = wx.App.Get()
    if isinstance(app, AmuletApp):
        return app
    else:
        raise Exception("wx App is not an instance of AmuletApp")


def open_level(path: str):
    """
    Open a level and create a tab for it.
    If a tab already exists it will just be shown.

    :param path: The path to the level to open.
    """
    get_app().open_level(path)


def close_level(path: str):
    """
    Close a level tab.

    :param path: The path to the level to close.
    """
    get_app().close_level(path)



import wx
from typing import Tuple

import PyMCTranslate

from amulet_map_editor.api.wx.ui.base_select import BaseSelect


class BlockSelect(BaseSelect):
    TypeName = "Block"

    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        platform: str,
        version_number: Tuple[int, int, int],
        force_blockstate: bool,
        namespace: str = None,
        block_name: str = None,
        show_pick_block: bool = False,
    ):
        super().__init__(
            parent,
            translation_manager,
            platform,
            version_number,
            force_blockstate,
            namespace,
            block_name,
            show_pick_block,
        )

    def _populate_namespace(self):
        version = self._translation_manager.get_version(
            self._platform, self._version_number
        )
        namespaces = version.block.namespaces(self._force_blockstate)
        self._do_text_event = False
        self._namespace_combo.Set(namespaces)

    def _populate_item_name(self):
        version = self._translation_manager.get_version(
            self._platform, self._version_number
        )
        self._names = version.block.base_names(self.namespace, self._force_blockstate)
        self._list_box.SetItems(self._names)


if __name__ == "__main__":

    def main():
        app = wx.App()
        translation_manager = PyMCTranslate.new_translation_manager()
        dialog = wx.Dialog(None)
        sizer = wx.BoxSizer()
        dialog.SetSizer(sizer)
        sizer.Add(BlockSelect(dialog, translation_manager, "java", (1, 16, 0), False))
        dialog.Show()
        dialog.Fit()
        app.MainLoop()

    main()



import os
import json

colours_path = os.path.join(os.path.dirname(__file__), "colours.json")

if os.path.isfile(colours_path):
    with open(colours_path) as f:
        colours = json.load(f)
    if isinstance(colours, dict):
        colours = {
            key: tuple(float(c) for c in val)
            for key, val in colours.items()
            if isinstance(val, list) and all(isinstance(c, (float, int)) for c in val)
        }
    else:
        colours = {}
else:
    colours = {}



import wx
from wx.lib import newevent
from typing import Tuple, List, Optional

import PyMCTranslate

from amulet_map_editor.api.image import COLOUR_PICKER

(
    ItemNamespaceChangeEvent,
    EVT_ITEM_NAMESPACE_CHANGE,
) = newevent.NewCommandEvent()  # the namespace entry changed
(
    ItemNameChangeEvent,
    EVT_ITEM_NAME_CHANGE,
) = newevent.NewCommandEvent()  # the name entry changed
(
    ItemChangeEvent,
    EVT_ITEM_CHANGE,
) = (
    newevent.NewCommandEvent()
)  # the name or namespace changed. Generated after EVT_ITEM_NAME_CHANGE
(
    PickEvent,
    EVT_PICK,
) = newevent.NewCommandEvent()  # The pick button was pressed


class BaseSelect(wx.Panel):
    TypeName = "?"

    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        platform: str,
        version_number: Tuple[int, int, int],
        force_blockstate: bool = None,
        namespace: str = None,
        default_name: str = None,
        show_pick: bool = False,
    ):
        super().__init__(parent, style=wx.BORDER_SIMPLE)
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        self._translation_manager = translation_manager

        self._platform: Optional[str] = None
        self._version_number: Optional[Tuple[int, int, int]] = None
        self._force_blockstate: Optional[bool] = force_blockstate

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._sizer.Add(sizer, 0, wx.EXPAND | wx.ALL, 5)
        text = wx.StaticText(self, label="Namespace:", style=wx.ALIGN_CENTER)
        sizer.Add(text, 1, wx.ALIGN_CENTER_VERTICAL)
        self._namespace_combo = wx.ComboBox(self)
        sizer.Add(self._namespace_combo, 2)
        self._set_version((platform, version_number, force_blockstate or False))
        self._populate_namespace()
        self.set_namespace(namespace)

        self._namespace_combo.Bind(
            wx.EVT_TEXT, lambda evt: self._post_namespace_change()
        )
        self._do_text_event = (
            True  # some widgets create events. This is used to suppress them
        )

        self.Bind(EVT_ITEM_NAMESPACE_CHANGE, self._on_namespace_change)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self._sizer.Add(sizer, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5)
        header_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(header_sizer, 0, wx.EXPAND | wx.BOTTOM, 5)
        header_sizer.Add(
            wx.StaticText(
                self, label=f"{self.TypeName.capitalize()} name:", style=wx.ALIGN_CENTER
            ),
            1,
            wx.ALIGN_CENTER_VERTICAL,
        )
        search_sizer = wx.BoxSizer(wx.HORIZONTAL)
        header_sizer.Add(search_sizer, 2, wx.EXPAND)
        self._search = wx.SearchCtrl(self)
        search_sizer.Add(self._search, 1, wx.ALIGN_CENTER_VERTICAL)
        self._search.Bind(wx.EVT_TEXT, self._on_search_change)
        if show_pick:
            pick_button = wx.BitmapButton(self, bitmap=COLOUR_PICKER.bitmap(22, 22))
            search_sizer.Add(pick_button, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)
            pick_button.Bind(
                wx.EVT_BUTTON,
                lambda evt: wx.PostEvent(self, PickEvent(self.GetId(), widget=self)),
            )
        self._list_box = wx.ListBox(self, style=wx.LB_SINGLE)
        sizer.Add(self._list_box, 1, wx.EXPAND)

        self._names: List[str] = []
        self._populate_item_name()
        self.set_name(default_name)
        self._list_box.Bind(wx.EVT_LISTBOX, lambda evt: self._post_item_change())

    def _post_namespace_change(self):
        if self._do_text_event:
            wx.PostEvent(
                self, ItemNamespaceChangeEvent(self.GetId(), namespace=self.namespace)
            )
        self._do_text_event = True

    def _post_item_change(self):
        wx.PostEvent(self, ItemNameChangeEvent(self.GetId(), name=self.name)),
        wx.PostEvent(
            self,
            ItemChangeEvent(self.GetId(), namespace=self.namespace, name=self.name),
        )

    @property
    def version(self) -> Tuple[str, Tuple[int, int, int], bool]:
        return self._platform, self._version_number, self._force_blockstate

    @version.setter
    def version(self, version: Tuple[str, Tuple[int, int, int], bool]):
        self._set_version(version)
        self._populate_namespace()
        self.namespace = None

    def _set_version(self, version: Tuple[str, Tuple[int, int, int], bool]):
        assert (
            version[0] in self._translation_manager.platforms()
            and version[1] in self._translation_manager.version_numbers(version[0])
            and isinstance(version[2], bool)
        ), f"{version} is not a valid version"
        self._platform, self._version_number, self._force_blockstate = version

    @property
    def namespace(self) -> str:
        return self._namespace_combo.GetValue()

    @namespace.setter
    def namespace(self, namespace: str):
        self.set_namespace(namespace)
        wx.PostEvent(
            self, ItemNamespaceChangeEvent(self.GetId(), namespace=self.namespace)
        )

    def set_namespace(self, namespace: str):
        namespace = namespace or "minecraft"
        if isinstance(namespace, str):
            if namespace in self._namespace_combo.GetItems():
                self._namespace_combo.SetSelection(
                    self._namespace_combo.GetItems().index(namespace)
                )
            else:
                self._namespace_combo.ChangeValue(namespace)

    @property
    def name(self) -> str:
        name: str = self._list_box.GetString(self._list_box.GetSelection())
        if self._list_box.GetSelection() == 0 and name.startswith('"'):
            name = name[1:-1]
        return name

    @name.setter
    def name(self, name: str):
        if self.set_name(name):
            self._post_item_change()

    def set_name(self, name: str) -> bool:
        name = name or ""
        self._search.ChangeValue(name)
        return self._update_item_name(name)

    def _populate_namespace(self):
        raise NotImplementedError("This method should be overridden in child classes.")

    def _populate_item_name(self):
        raise NotImplementedError("This method should be overridden in child classes.")

    def _on_namespace_change(self, evt):
        self._populate_item_name()
        self.name = None
        evt.Skip()

    def _on_search_change(self, evt):
        search_str = evt.GetString()
        if self._update_item_name(search_str):
            self._post_item_change()

    def _update_item_name(self, search_str: str) -> bool:
        names = [bn for bn in self._names if search_str in bn]
        if search_str not in names:
            names.insert(0, f'"{search_str}"')

        index = 0
        selection = self._list_box.GetSelection()
        if selection != wx.NOT_FOUND:
            current_string = self._list_box.GetString(selection)
            if current_string in names:
                index = names.index(current_string)

        self._list_box.SetItems(names)
        if index:
            # if the previously selected string is in the list select that
            self._list_box.SetSelection(index)
            return False
        elif search_str in names:
            # if the searched text perfectly matches select that
            self._list_box.SetSelection(names.index(search_str))
            return True
        elif len(self._list_box.GetItems()) >= 2:
            self._list_box.SetSelection(1)
            return True
        else:
            self._list_box.SetSelection(0)
            return True



import os
import pkgutil
import sys
from typing import Optional
from types import ModuleType
import importlib.util

from amulet_map_editor.programs.edit.plugins.operations import stock_plugins


STOCK_PLUGINS_DIR: str = stock_plugins.__path__[0]
STOCK_PLUGINS_NAME: str = stock_plugins.__name__
CUSTOM_PLUGINS_DIR = os.path.abspath("plugins")



from amulet_map_editor.programs.edit.api.ui.tool import BaseOperationChoiceToolUI


class ExportTool(BaseOperationChoiceToolUI):
    OperationGroupName = "export_operations"
    ShowOpenFolder = False

    @property
    def name(self) -> str:
        return "Export"



class BaseOperationException(Exception):
    pass


class BaseLoudException(BaseOperationException):
    pass


class BaseSilentException(BaseOperationException):
    pass


class OperationError(BaseLoudException):
    """Error to raise if something went wrong when running the operation.
    Will notify the user with a dialog box with the message f"Error running operation {msg}"
    Changes will be rolled back to the last backup.
    Eg. if the operation requires something it is not given"""

    pass


class OperationSuccessful(BaseLoudException):
    """raise this if you want to exit the operation without creating an undo point.
    Will notify the user with a dialog box containing the message.
    Changes will be rolled back to the last backup."""

    pass


class OperationSilentAbort(BaseSilentException):
    """Error to raise if something went wrong when running the operation but you do not want the user to see.
    This error will be handled silently.
    Changes will be rolled back to the last backup."""

    pass



"""
This license applies to this file only.
-- begin license --
MIT License
Copyright (c) 2021 Amulet-Team
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-- end license --
"""

from typing import TYPE_CHECKING, Tuple
import wx
import numpy

from amulet.api.block import Block

from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.block_select import BlockDefine
from amulet_map_editor.api.wx.ui.simple import SimpleScrollablePanel
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Replace(SimpleScrollablePanel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleScrollablePanel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        options = self._load_options({})

        self._original_block = BlockDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("original_block_options", [])
                or [world.level_wrapper.platform]
            ),
            wildcard_properties=True,
            show_pick_block=True
        )
        self._sizer.Add(self._original_block, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)
        self._original_block.Bind(EVT_PICK, lambda evt: self._on_pick_block_button(1))
        self._replacement_block = BlockDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("replacement_block_options", [])
                or [world.level_wrapper.platform]
            ),
            show_pick_block=True
        )
        self._sizer.Add(
            self._replacement_block, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5
        )
        self._replacement_block.Bind(
            EVT_PICK, lambda evt: self._on_pick_block_button(2)
        )

        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self.Layout()
        self.Thaw()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        return (1,)

    def _on_pick_block_button(self, code):
        """Set up listening for the block click"""
        self._show_pointer = code

    def _on_box_click(self):
        if self._show_pointer:
            x, y, z = self._pointer.pointer_base
            if self._show_pointer == 1:
                self._original_block.universal_block = (
                    self.world.get_block(x, y, z, self.canvas.dimension),
                    None,
                )
            elif self._show_pointer == 2:
                self._replacement_block.universal_block = (
                    self.world.get_block(x, y, z, self.canvas.dimension),
                    None,
                )
            self._show_pointer = False

    def _get_replacement_block(self) -> Block:
        return self._replacement_block.universal_block[0]

    def disable(self):
        self._save_options(
            {
                "original_block_options": (
                    self._original_block.platform,
                    self._original_block.version_number,
                    self._original_block.force_blockstate,
                    self._original_block.namespace,
                    self._original_block.block_name,
                    self._original_block.str_properties,
                ),
                "replacement_block": self._get_replacement_block(),
                "replacement_block_options": (
                    self._replacement_block.platform,
                    self._replacement_block.version_number,
                    self._replacement_block.force_blockstate,
                    self._replacement_block.namespace,
                    self._replacement_block.block_name,
                    self._replacement_block.str_properties,
                ),
            }
        )

    def _run_operation(self, _):
        self.canvas.run_operation(lambda: self._replace())

    def _replace(self):
        world = self.world
        selection = self.canvas.selection.selection_group
        dimension = self.canvas.dimension

        (
            original_platform,
            original_version,
            original_blockstate,
            original_namespace,
            original_base_name,
            original_properties,
        ) = (
            self._original_block.platform,
            self._original_block.version_number,
            self._original_block.force_blockstate,
            self._original_block.namespace,
            self._original_block.block_name,
            self._original_block.str_properties,
        )
        replacement_block = self._get_replacement_block()

        replacement_block_id = world.block_palette.get_add_block(replacement_block)

        original_block_matches = []
        universal_block_count = 0

        iter_count = len(list(world.get_coord_box(dimension, selection)))
        count = 0

        for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection):
            if universal_block_count < len(world.block_palette):
                for universal_block_id in range(
                    universal_block_count, len(world.block_palette)
                ):
                    version_block = world.translation_manager.get_version(
                        original_platform, original_version
                    ).block.from_universal(
                        world.block_palette[universal_block_id],
                        force_blockstate=original_blockstate,
                    )[
                        0
                    ]
                    if (
                        version_block.namespace == original_namespace
                        and version_block.base_name == original_base_name
                        and all(
                            original_properties.get(prop) in ["*", val.to_snbt()]
                            for prop, val in version_block.properties.items()
                        )
                    ):
                        original_block_matches.append(universal_block_id)

                universal_block_count = len(world.block_palette)
            blocks = chunk.blocks[slices]
            blocks[numpy.isin(blocks, original_block_matches)] = replacement_block_id
            chunk.blocks[slices] = blocks
            chunk.changed = True

            count += 1
            yield count / iter_count

    def DoGetBestClientSize(self):
        sizer = self.GetSizer()
        if sizer is None:
            return -1, -1
        else:
            sx, sy = self.GetSizer().CalcMin()
            return (
                sx + wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X),
                sy + wx.SystemSettings.GetMetric(wx.SYS_HSCROLL_Y),
            )


export = {
    "name": "Replace",  # the name of the plugin
    "operation": Replace,  # the actual function to call when running the plugin
}



import wx

from amulet.api.level import BaseLevel
from amulet_map_editor.api.datatypes import MenuData
from amulet_map_editor.api.framework.base_tab import BaseTab


class BaseProgram(BaseTab):
    def __init__(self, container: wx.Window, world: BaseLevel):
        pass

    def menu(self, menu: MenuData) -> MenuData:
        return menu



import time
from typing import Optional
from threading import Thread

from amulet_map_editor.api.opengl import ThreadedObjectContainer

ThreadingEnabled = True
if ThreadingEnabled:

    class ChunkGenerator(ThreadedObjectContainer):
        def __init__(self):
            ThreadedObjectContainer.__init__(self)
            self._enabled = False
            self._thread: Optional[Thread] = None

        def start(self):
            if not self._enabled:
                if self._thread is not None:
                    raise Exception("Thread being disabled")
                self._enabled = True
                self._thread = Thread(target=self._generate_chunks)
                self._thread.start()

        def stop(self):
            if self._enabled:
                self._enabled = False
                self._thread.join()
                self._thread = None

        def _generate_chunks(self):
            while self._enabled:
                self.thread_action()
                # we need to force the chunk generator to share with the UI code otherwise it eats up all processing time.
                # TODO: find a better way to do this
                time.sleep(0.01)

else:

    class ChunkGenerator(ThreadedObjectContainer):
        def start(self):
            pass

        def stop(self):
            pass



from typing import List, Union
import numpy
from .render_selection_group import (
    RenderSelectionGroup,
)
from amulet_map_editor.api.opengl.mesh.selection import RenderSelectionHighlightable


class RenderSelectionGroupHighlightable(RenderSelectionGroup):
    """A group of selection boxes to be drawn with an added editable box."""

    def _new_render_selection(self):
        return RenderSelectionHighlightable(
            self._context_identifier, self.resource_pack
        )

    def reset_highlight_edges(self):
        self._boxes: List[RenderSelectionHighlightable]
        for box in self._boxes:
            box.reset_highlight_edges()

    def set_highlight_edges(
        self, box_index: int, highlight_edges: Union[numpy.ndarray, bool]
    ):
        self._boxes: List[RenderSelectionHighlightable]
        self._boxes[box_index].set_highlight_edges(highlight_edges)



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.selection import SelectionGroup
from amulet.api.block import UniversalAirBlock
from amulet.api.data_types import Dimension, OperationReturnType

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def delete(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
) -> OperationReturnType:
    internal_id = world.block_palette.get_add_block(UniversalAirBlock)

    iter_count = len(list(world.get_coord_box(dimension, selection, False)))
    count = 0

    for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection, False):
        chunk.blocks[slices] = internal_id
        chunk.changed = True
        count += 1
        yield count / iter_count



import numpy
from typing import List

from amulet_map_editor.api.opengl import Drawable, ContextManager
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePack,
    OpenGLResourcePackManagerStatic,
)
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.data_types import PointCoordinatesAny
from amulet_map_editor.api.opengl.mesh.selection import RenderSelection


class RenderSelectionGroup(Drawable, ContextManager, OpenGLResourcePackManagerStatic):
    """A group of selection boxes to be drawn"""

    def __init__(
        self,
        context_identifier: str,
        resource_pack: OpenGLResourcePack,
        selection: SelectionGroup = None,
    ):
        ContextManager.__init__(self, context_identifier)
        OpenGLResourcePackManagerStatic.__init__(self, resource_pack)

        self._boxes: List[RenderSelection] = []

        if selection:
            self.selection_group = selection

    def __len__(self):
        return len(self._boxes)

    def __bool__(self):
        return bool(self._boxes)

    def _new_render_selection(self):
        return RenderSelection(self.context_identifier, self.resource_pack)

    @property
    def selection_group(self) -> SelectionGroup:
        return SelectionGroup(
            [SelectionBox(box.point1, box.point2) for box in self._boxes]
        )

    @selection_group.setter
    def selection_group(self, selection_group: SelectionGroup):
        if not isinstance(selection_group, SelectionGroup):
            raise TypeError("selection_group must be a SelectionGroup.")
        self.unload()
        for box in selection_group.selection_boxes:
            render_box = self._new_render_selection()
            render_box.point1 = numpy.array(box.point_1)
            render_box.point2 = numpy.array(box.point_2)
            self._boxes.append(render_box)

    def draw(
        self, camera_matrix: numpy.ndarray, camera_position: PointCoordinatesAny = None
    ):
        for box in self._boxes:
            box.draw(camera_matrix, camera_position)

    def unload(self):
        while self._boxes:
            box = self._boxes.pop()
            box.unload()



from OpenGL.GL import (
    GL_DYNAMIC_DRAW,
    glBindVertexArray,
    glBindBuffer,
    GL_ARRAY_BUFFER,
    glBufferSubData,
)
from typing import Dict, Tuple, Optional
import numpy
import queue
from .chunk import RenderChunk
from amulet_map_editor.api.opengl.mesh.tri_mesh import TriMesh
from amulet_map_editor.api.opengl.resource_pack import OpenGLResourcePack
from amulet_map_editor.api.opengl.matrix import displacement_matrix
from amulet_map_editor.api.opengl.data_types import TransformationMatrix


class ChunkManager:
    def __init__(
        self, context_identifier: str, resource_pack: OpenGLResourcePack, region_size=16
    ):
        self.context_identifier = context_identifier
        self._resource_pack = resource_pack
        self.region_size = region_size
        self._regions: Dict[Tuple[int, int], RenderRegion] = {}
        # added chunks are put in here and then processed on the next call of draw
        # This is because add_render_chunk can be called from a different thread to draw
        # which causes issues due to dictionaries resizing
        self._chunk_temp: queue.Queue = queue.Queue()
        self._chunk_temp_set = set()
        self._rebuild_regions = []

    def add_render_chunk(self, render_chunk: RenderChunk):
        """Add a RenderChunk to the database.
        A call to _merge_chunk_temp from the main thread will be needed for them to be drawn.
        This is done after the next draw call."""
        self._chunk_temp.put(render_chunk)
        chunk_coords = (render_chunk.cx, render_chunk.cz)
        self._chunk_temp_set.add(chunk_coords)

    def render_chunk_needs_rebuild(self, chunk_coords: Tuple[int, int]) -> bool:
        return (
            chunk_coords not in self._chunk_temp_set
            and self.render_chunk_in_main_database(chunk_coords)
            and self.get_render_chunk(chunk_coords).needs_rebuild()
        )

    def get_render_chunk(self, chunk_coords: Tuple[int, int]) -> RenderChunk:
        """Get a RenderChunk from the database.
        Might throw a key error if it has not been added to the real database yet."""
        return self._regions[self.region_coords(*chunk_coords)].get_render_chunk(
            chunk_coords
        )

    def _merge_chunk_temp(self):
        for _ in range(self._chunk_temp.qsize()):
            render_chunk = self._chunk_temp.get()
            region_coords = self.region_coords(render_chunk.cx, render_chunk.cz)
            if region_coords not in self._regions:
                self._regions[region_coords] = RenderRegion(
                    region_coords[0],
                    region_coords[1],
                    self.region_size,
                    self.context_identifier,
                    self._resource_pack,
                )
            self._regions[region_coords].add_render_chunk(render_chunk)
        self._chunk_temp_set.clear()

    def __contains__(self, chunk_coords: Tuple[int, int]):
        return (
            chunk_coords in self._chunk_temp_set
            or self.render_chunk_in_main_database(chunk_coords)
        )

    def render_chunk_in_main_database(self, chunk_coords: Tuple[int, int]) -> bool:
        region_coords = self.region_coords(*chunk_coords)
        return (
            region_coords in self._regions
            and chunk_coords in self._regions[region_coords]
        )

    def region_coords(self, cx, cz):
        return cx // self.region_size, cz // self.region_size

    def draw(self, camera_matrix: TransformationMatrix, camera):
        cam_rx, cam_rz = numpy.floor(
            numpy.array(camera)[[0, 2]] / (16 * self.region_size)
        )
        cam_cx, cam_cz = numpy.floor(numpy.array(camera)[[0, 2]] / 16)
        for region in sorted(
            self._regions.values(),
            key=lambda x: abs(x.rx - cam_rx) + abs(x.rz - cam_rz),
            reverse=True,
        ):
            region.draw(camera_matrix, cam_cx, cam_cz)
        self._merge_chunk_temp()

    def unload(self, safe_area: Tuple[int, int, int, int] = None):
        if safe_area is None:
            for _ in range(self._chunk_temp.qsize()):
                self._chunk_temp.get()
            self._chunk_temp_set.clear()
            for region in self._regions.values():
                region.unload()
            self._regions.clear()
        else:
            min_rx, min_rz = self.region_coords(*safe_area[:2])
            max_rx, max_rz = self.region_coords(*safe_area[2:])
            delete_regions = []
            for region in self._regions.values():
                if not (
                    min_rx <= region.rx <= max_rx and min_rz <= region.rz <= max_rz
                ):
                    region.unload()
                    delete_regions.append((region.rx, region.rz))

            for region in delete_regions:
                del self._regions[region]

    def rebuild(self):
        """Rebuild a single region which was last rebuild the longest ago.
        Put this on a semi-fast clock to rebuild all regions."""
        if not self._rebuild_regions:
            self._rebuild_regions = list(self._regions.keys())

        if self._rebuild_regions:
            region = self._rebuild_regions.pop(0)
            if region in self._regions:
                self._regions[region].rebuild()


MergedChunkLocationsType = Dict[Tuple[int, int], Tuple[int, int, int, int]]


class RenderRegion(TriMesh):
    _merged_chunk_locations: MergedChunkLocationsType
    _temp_data: Optional[Tuple[numpy.ndarray, MergedChunkLocationsType]]

    def __init__(
        self,
        rx: int,
        rz: int,
        region_size: int,
        context_identifier: str,
        resource_pack: OpenGLResourcePack,
    ):
        """A group of RenderChunks to minimise the number of draw calls"""
        super().__init__(context_identifier, resource_pack)
        self.rx = rx
        self.rz = rz
        self._chunks: Dict[Tuple[int, int], RenderChunk] = {}
        self._merged_chunk_locations: MergedChunkLocationsType = {}
        self._manual_chunks: Dict[Tuple[int, int], RenderChunk] = {}

        # Merging is done on a new thread which can't modify the opengl state.
        # This stores the created data and the main thread loads it when drawing.
        self._temp_data = None

        self.region_transform = displacement_matrix(
            rx * region_size * 16, 0, rz * region_size * 16
        )

    @property
    def vertex_usage(self):
        return GL_DYNAMIC_DRAW

    def __repr__(self):
        return f"RenderRegion({self.rx}, {self.rz})"

    def __contains__(self, item):
        return item in self._chunks

    def add_render_chunk(self, render_chunk: RenderChunk):
        """Add a chunk to the region"""
        chunk_coords = (render_chunk.cx, render_chunk.cz)
        if chunk_coords in self._chunks:
            self._chunks[chunk_coords].unload()
        self._disable_merged_chunk(chunk_coords)
        self._chunks[chunk_coords] = render_chunk
        self._manual_chunks[chunk_coords] = render_chunk

    def get_render_chunk(self, chunk_coords: Tuple[int, int]):
        return self._chunks[chunk_coords]

    def _disable_merged_chunk(self, chunk_coords: Tuple[int, int]):
        """Zero out the region of memory in the merged chunks related to a given chunk"""
        if chunk_coords in self._merged_chunk_locations:
            (
                offset,
                size,
                translucent_offset,
                translucent_size,
            ) = self._merged_chunk_locations.pop(chunk_coords)
            glBindVertexArray(self._vao)
            glBindBuffer(GL_ARRAY_BUFFER, self._vbo)
            glBufferSubData(
                GL_ARRAY_BUFFER,
                offset * 4,
                size * 4,
                numpy.zeros(size, dtype=numpy.float32),
            )
            glBufferSubData(
                GL_ARRAY_BUFFER,
                translucent_offset * 4,
                translucent_size * 4,
                numpy.zeros(translucent_size, dtype=numpy.float32),
            )
            glBindVertexArray(0)
            glBindBuffer(GL_ARRAY_BUFFER, 0)

    def rebuild(self):
        """Merges chunk geometry for the region into one large array.
        As each chunk is added it is drawn individually.
        After not too long the individual draw calls for each chunk will reach the bottleneck for python.
        To solve this we take the geometry for each chunk and merge them into one large array that only requires one draw call.

        :return:
        """
        if self._manual_chunks:
            region_verts = []
            region_verts_translucent = []
            merged_locations: MergedChunkLocationsType = {}
            offset = 0
            translucent_offset = 0
            for chunk_location, chunk in self._chunks.items():
                region_verts.append(chunk.verts[: chunk.verts_translucent])
                region_verts_translucent.append(chunk.verts[chunk.verts_translucent :])
                merged_locations[chunk_location] = [
                    offset,
                    chunk.verts_translucent,
                    translucent_offset,
                    chunk.verts.size - chunk.verts_translucent,
                ]
                offset += chunk.verts_translucent
                translucent_offset += chunk.verts.size - chunk.verts_translucent

            for val in merged_locations.values():
                val[2] += offset

            region_verts += region_verts_translucent

            if region_verts:
                verts = numpy.concatenate(region_verts)
            else:
                verts = self.new_empty_verts()
            self._temp_data = verts, merged_locations

    def _create_geometry(self):
        """Load the temporary vertex data into opengl."""
        if self._temp_data is not None:
            self._setup()
            verts, merged_locations = self._temp_data
            self._temp_data = None
            self.draw_count = int(verts.size // self._vert_len)
            self._merged_chunk_locations = merged_locations

            self.change_verts(verts)
            for coord in merged_locations:
                chunk = self._manual_chunks.pop(coord, None)
                if chunk is not None:
                    chunk.unload()

    def unload(self):
        """Unload all opengl data"""
        super().unload()
        for chunk in self._chunks.values():
            chunk.unload()
        self._chunks.clear()

    def draw(self, camera_matrix: TransformationMatrix, cam_cx, cam_cz):
        self._create_geometry()
        transformation_matrix = numpy.matmul(camera_matrix, self.region_transform)
        super().draw(transformation_matrix)
        for chunk in sorted(
            self._manual_chunks.values(),
            key=lambda x: abs(x.cx - cam_cx) + abs(x.cz - cam_cz),
            reverse=True,
        ):
            chunk.draw(transformation_matrix)



import wx
from amulet_map_editor import lang


class WarningDialog(wx.Dialog):
    def __init__(self, parent: wx.Window):
        super().__init__(parent, style=wx.CAPTION)
        self.SetTitle(lang.get("warning_dialog.title"))

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        content = wx.StaticText(
            self,
            wx.ID_ANY,
            lang.get("warning_dialog.content"),
            style=wx.ALIGN_CENTER_HORIZONTAL,
        )
        content.Wrap(750)
        main_sizer.Add(content, 1, wx.ALL | wx.EXPAND, 5)

        button_sizer = wx.StdDialogButtonSizer()
        main_sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 4)

        self._do_not_show = wx.CheckBox(
            self, wx.ID_ANY, lang.get("warning_dialog.do_not_show_again")
        )
        button_sizer.Add(self._do_not_show, 0, wx.ALIGN_CENTER_VERTICAL, 0)

        self._understand_button = wx.Button(
            self, wx.ID_ANY, lang.get("warning_dialog.i_understand")
        )
        self._understand_button.SetDefault()
        button_sizer.Add(self._understand_button, 0, 0, 0)

        button_sizer.Realize()

        self.SetSizer(main_sizer)
        main_sizer.Fit(self)

        self.SetAffirmativeId(self._understand_button.GetId())

        self.Layout()

    @property
    def do_not_show_again(self) -> bool:
        return self._do_not_show.GetValue()



import wx
from typing import List, Tuple, Any
import logging

from amulet_map_editor.api.opengl.canvas import BaseCanvas

log = logging.getLogger(__name__)


class EventCanvas(BaseCanvas):
    """A modification of the normal canvas to make it easier to add and remove events."""

    def __init__(self, parent: wx.Window):
        self._bound_events: List[Tuple[wx.PyEventBinder, Any, Any]] = []
        super().__init__(parent)

    def reset_bound_events(self):
        """Unbind all events and re-bind the default events.
        We are allowing users to bind custom events so we should have a way to reset what is bound."""
        self.tear_down_events()
        self.bind_events()

    def tear_down_events(self):
        """Unbind all events.
        We are allowing users to bind custom events so we should have a way to reset what is bound."""
        for event, handler, source in self._bound_events:
            if source is None:
                while super().Unbind(event):
                    pass
            else:
                if not self.Unbind(event, source, handler=handler):
                    log.error(f"Failed to unbind {event}, {handler}, {source}")
        self._bound_events.clear()

    def bind_events(self):
        """Set up all events required to run.
        Note this will also bind subclass events."""
        raise NotImplementedError

    def Bind(self, event, handler, source=None, id=wx.ID_ANY, id2=wx.ID_ANY):
        """Bind an event to the canvas."""
        self._bound_events.append((event, handler, source))
        super().Bind(event, handler, source, id, id2)

    def Unbind(
        self, event, source=None, id=wx.ID_ANY, id2=wx.ID_ANY, handler=None
    ) -> bool:
        """Unbind an event from the canvas."""
        key = (event, handler, source)
        if key in self._bound_events:
            self._bound_events.remove(key)
        return super().Unbind(event, source=source, id=id, id2=id2, handler=handler)



import numpy

from minecraft_model_reader.api.mesh.block.cube import get_cube

from .tri_mesh import TriMesh
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManager,
    OpenGLResourcePack,
)
from amulet_map_editor.api.opengl.data_types import CameraLocationType
from amulet_map_editor.api.opengl.matrix import displacement_matrix

Radius = 2000


class SkyBox(TriMesh, OpenGLResourcePackManager):
    """A class to define the logic to generate geometry from a block array"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        TriMesh.__init__(self, context_identifier, resource_pack)
        OpenGLResourcePackManager.__init__(self, resource_pack)
        self._camera_location: CameraLocationType = (0, 0, 0)
        self._rebuild()

    def _rebuild(self):
        model = get_cube(
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/down"),
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/up"),
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/north"),
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/west"),
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/south"),
            self.resource_pack.get_texture_path("amulet", "amulet_ui/cubemap/east"),
            bounds=((Radius, -Radius), (-Radius, Radius), (-Radius, Radius)),
            do_not_cull=(True,) * 6,
        )

        verts = model.verts[None].reshape((-1, 3))
        tverts = model.texture_coords[None].reshape((-1, 2))
        faces = model.faces[None]

        # each slice in the first axis is a new block, each slice in the second is a new vertex
        vert_table = numpy.zeros((faces.size, self._vert_len), dtype=numpy.float32)
        vert_table[:, :3] = verts[faces]
        vert_table[:, 3:5] = tverts[faces]

        vert_index = 0
        for texture_index in model.texture_index[None]:
            tex_bounds = self.resource_pack.texture_bounds(
                model.textures[texture_index]
            )

            vert_table[vert_index : vert_index + 3, 5:9] = tex_bounds
            vert_index += 3

        vert_table[:, 9:12] = model.tint_verts[None].reshape((-1, 3))[faces]
        self.verts = numpy.ravel(vert_table)
        self.draw_count = int(self.verts.size // self._vert_len)

    def set_camera_location(self, camera_location: CameraLocationType):
        self._camera_location = (numpy.asarray(camera_location)).tolist()

    def draw(self, transformation_matrix: numpy.ndarray):
        super().draw(
            numpy.matmul(
                transformation_matrix, displacement_matrix(*self._camera_location)
            )
        )



# Fixed function example plugin 2
# see example 1 before looking at this example
# Notes about the operation


from amulet.api.selection import SelectionGroup
from amulet.api.level import BaseLevel
from amulet.api.data_types import Dimension


# Notes about the operation
# The operation is allowed to yield floats in the range 0 to 1.
#       This is used to update the loading bar in the UI. Without this the UI may appear to be not responding
#       It can optionally also yield a float and a string. The float is the same as the above and the string is used to display in the loading bar
# The operation is allowed to return a value however nothing will be done with it


def operation(
    world: BaseLevel, dimension: Dimension, selection: SelectionGroup, options: dict
):
    for i in range(10):
        # do some logic
        yield (i + 1) / 10

    for i in range(10):
        # do some logic
        yield (i + 1) / 10, f"Step {i} of 10"
    return "hello"  # This will not actually do anything but is allowed


export = {
    "name": "Fixed Function Pipeline Example 2",
    "operation": operation,
}



#!/usr/bin/env python3


def _on_error(e):
    """Code to handle errors"""
    err_list = []
    try:
        import traceback

        err_list.append(traceback.format_exc())
    except:
        pass
    if isinstance(e, ImportError):
        err_list.append(
            f"Failed to import requirements. Check that you extracted correctly."
        )
    err_list.append(str(e))
    err = "\n".join(err_list)
    print(err)
    with open("./logs/crash.log", "w") as f:
        f.write(err)
    input("Press ENTER to continue.")
    sys.exit(1)


try:
    import sys

    if sys.version_info[:2] < (3, 7):
        raise Exception("Must be using Python 3.7+")
    import logging
    import os
    import traceback
    import glob
    import time
    import wx

    if sys.platform == "linux" and wx.VERSION >= (4, 1, 1):
        # bug 247
        os.environ["PYOPENGL_PLATFORM"] = "egl"
except Exception as e:
    _on_error(e)


def _init_log():
    logs_path = os.path.join(".", "logs")
    # set up handlers
    os.makedirs(logs_path, exist_ok=True)
    # remove all log files older than a week
    for path in glob.glob(os.path.join(logs_path, "*.log")):
        if (
            os.path.isfile(path)
            and os.path.getmtime(path) < time.time() - 3600 * 24 * 7
        ):
            os.remove(path)

    log = logging.getLogger()
    log.setLevel(logging.DEBUG if "amulet-debug" in sys.argv else logging.INFO)

    file_handler = logging.FileHandler(
        os.path.join(logs_path, f"amulet_{os.getpid()}.log"), "w", encoding="utf-8"
    )
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    log.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    log.addHandler(console_handler)


def main():
    try:
        _init_log()
        from amulet_map_editor.api.framework import AmuletApp

    except Exception as e:
        _on_error(e)
        AmuletApp = None

    try:
        app = AmuletApp(0)
        app.MainLoop()
    except Exception as e:
        log = logging.getLogger(__name__)
        log.critical(
            f"Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue. \n{traceback.format_exc()}"
        )
        input("Press ENTER to continue.")
    sys.exit(0)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from amulet.api.structure import structure_cache
from amulet.api.data_types import Dimension, OperationReturnType
from amulet.api.selection import SelectionGroup
from amulet_map_editor.programs.edit.plugins.operations.stock_plugins.internal_operations.delete import (
    delete,
)
from amulet_map_editor.programs.edit.api.operations import OperationError

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def cut(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
) -> OperationReturnType:
    if selection:
        structure = world.extract_structure(selection, dimension)
        structure_cache.add_structure(structure, structure.dimensions[0])
        yield from delete(
            world,
            dimension,
            selection,
        )
    else:
        raise OperationError(
            "At least one selection is required for the copy operation."
        )



import wx
from typing import Union, Any

from amulet_map_editor.programs.edit.api.edit_canvas_container import (
    EditCanvasContainer,
)
from ..canvas_toggle_element import CanvasToggleElement

BaseToolUIType = Union[wx.Window, wx.Sizer, "BaseToolUI"]


class BaseToolUI(EditCanvasContainer, CanvasToggleElement):
    """The abstract base class for all tools that are to be loaded into the canvas."""

    @property
    def name(self) -> str:
        """The name of the tool."""
        raise NotImplementedError

    def enable(self):
        """Set the state of the tool for being enabled.
        Do not bind events to the canvas here because they will get removed. Do this in bind_events."""
        pass

    def bind_events(self):
        """Bind all required events to the canvas.
        All events on the canvas will be automatically removed after the tool is disabled.
        """
        pass

    def set_state(self, state: Any):
        """
        Set any state data.
        In some cases one tool might bounce to another and want to do more than just starting it.
        This will get called after enabling the tool.

        :param state: The state to set. Validate that this data is correct because it may come from anywhere.
        """
        pass

    def disable(self):
        """Stop the tool. Unload any excessive data. May get resumed again with a call to enable.
        All events bound to the canvas will be automatically removed after this is run."""
        pass



import weakref
from wx.glcanvas import GLCanvas


class CanvasContainer:
    """A helper class to store a reference to a canvas.
    If a canvas is hard referenced there may be cyclic references leading to memory leaks.
    Subclass this class if you intend to store a reference to the canvas."""

    def __init__(self, canvas: GLCanvas):
        self._canvas = weakref.ref(canvas)

    @property
    def canvas(self) -> GLCanvas:
        return self._canvas()



import os
import wx
import glob
from sys import platform
from typing import List, Dict, Tuple, Callable, TYPE_CHECKING
import traceback
import logging

from amulet import load_format
from amulet.api.errors import FormatError

from amulet_map_editor import lang, CONFIG
from amulet_map_editor.api.wx.ui import simple
from amulet_map_editor.api.wx.util.ui_preferences import preserve_ui_preferences
from amulet_map_editor.api.framework import app


if TYPE_CHECKING:
    from amulet.api.wrapper import WorldFormatWrapper

log = logging.getLogger(__name__)


# Windows 	%APPDATA%\.minecraft
# macOS 	~/Library/Application Support/minecraft
# Linux 	~/.minecraft


def get_java_dir():
    if platform == "win32":
        return os.path.join(os.getenv("APPDATA"), ".minecraft")
    elif platform == "darwin":
        return os.path.expanduser("~/Library/Application Support/minecraft")
    else:
        return os.path.expanduser("~/.minecraft")


def get_java_saves_dir():
    return os.path.join(get_java_dir(), "saves")


minecraft_world_paths = {lang.get("world.java_platform"): get_java_saves_dir()}

if platform == "win32":
    minecraft_world_paths[lang.get("world.bedrock_platform")] = os.path.join(
        os.getenv("LOCALAPPDATA"),
        "Packages",
        "Microsoft.MinecraftUWP_8wekyb3d8bbwe",
        "LocalState",
        "games",
        "com.mojang",
        "minecraftWorlds",
    )

world_images: Dict[str, Tuple[int, wx.Bitmap, int]] = {}


def get_world_image(image_path: str) -> Tuple[wx.Bitmap, int]:
    if (
        image_path not in world_images
        or world_images[image_path][0] != os.stat(image_path)[8]
    ):
        img = wx.Image(image_path, wx.BITMAP_TYPE_ANY)
        width = min(int((img.GetWidth() / img.GetHeight()) * 128), 300)

        world_images[image_path] = (
            os.stat(image_path)[8],
            img.Scale(width, 128, wx.IMAGE_QUALITY_NEAREST).ConvertToBitmap(),
            width,
        )

    return world_images[image_path][1:3]


class WorldUI(wx.Panel):
    """A Panel UI element with the world image, name and description"""

    def __init__(self, parent: wx.Window, world_format: "WorldFormatWrapper"):
        super().__init__(parent)
        self.SetWindowStyle(wx.TAB_TRAVERSAL | wx.BORDER_RAISED)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(sizer)

        img, width = get_world_image(world_format.world_image_path)

        self.img = wx.StaticBitmap(self, wx.ID_ANY, img, (0, 0), (width, 128))
        sizer.Add(self.img)

        self.world_name = wx.StaticText(
            self,
            label="\n".join(
                [
                    world_format.level_name,
                    world_format.game_version_string,
                    os.path.join(
                        *os.path.normpath(world_format.path).split(os.sep)[-3:]
                    ),
                ]
            ),
        )
        sizer.Add(self.world_name, 0, wx.ALL | wx.ALIGN_CENTER, 5)


class WorldUIButton(WorldUI):
    """A Panel UI element that behaves like a button with the world image, name and description"""

    def __init__(
        self,
        parent: wx.Window,
        world_format: "WorldFormatWrapper",
        open_world_callback,
    ):
        super().__init__(parent, world_format)
        self.path = world_format.path
        self.open_world_callback = open_world_callback

        self.Bind(wx.EVT_LEFT_UP, self._call_callback)
        self.img.Bind(wx.EVT_LEFT_UP, self._call_callback)
        self.world_name.Bind(wx.EVT_LEFT_DOWN, self._call_callback)

    def _call_callback(self, evt):
        self.open_world_callback(self.path)


class WorldList(wx.Panel):
    """A Panel containing zero or more `WorldUIButton`s."""

    def __init__(self, parent: wx.Window, world_dirs, open_world_callback, sort=True):
        super().__init__(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self.worlds = []

        world_formats = []
        for world_path in world_dirs:
            if os.path.isdir(world_path):
                try:
                    world_formats.append(load_format(world_path))
                except FormatError as e:
                    log.info(f"Could not find loader for {world_path} {e}")
                except Exception:
                    log.error(
                        f"Error loading format wrapper for {world_path} {traceback.format_exc()}"
                    )
        if sort:
            world_formats = reversed(sorted(world_formats, key=lambda f: f.last_played))

        for world_format in world_formats:
            try:
                world_button = WorldUIButton(self, world_format, open_world_callback)
                sizer.Add(
                    world_button, 0, wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, 5
                )
                self.worlds.append(world_button)
            except Exception as e:
                log.info(f"Failed to display world button for {world_format.path} {e}")

        self.Layout()


class CollapsibleWorldListUI(wx.CollapsiblePane):
    """a drop down list of `WorldUIButton`s for a given directory"""

    def __init__(self, parent, paths: List[str], group_name: str, open_world_callback):
        super().__init__(parent, label=group_name)
        self.parent = parent
        self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, self.eval_layout)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        panel = self.GetPane()
        panel.sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(panel.sizer)
        panel.sizer.Add(WorldList(panel, paths, open_world_callback), 0, wx.EXPAND)

    def eval_layout(self, evt):
        self.Layout()
        self.parent.FitInside()
        evt.Skip()


class ScrollableWorldsUI(simple.SimpleScrollablePanel):
    # a frame to allow scrolling
    def __init__(self, parent, open_world_callback):
        super(ScrollableWorldsUI, self).__init__(parent)
        self.open_world_callback = open_world_callback

        self.dirs: Dict[str, CollapsibleWorldListUI] = {}
        self.reload()

        self.Layout()

    def reload(self):
        for val in self.dirs.values():
            val.Destroy()
        self.dirs.clear()
        for group_name, directory in minecraft_world_paths.items():
            if os.path.isdir(directory):
                world_list = CollapsibleWorldListUI(
                    self,
                    glob.glob(os.path.join(glob.escape(directory), "*")),
                    group_name,
                    self.open_world_callback,
                )
                self.add_object(world_list, 0, wx.EXPAND)
                self.dirs[directory] = world_list

    def OnChildFocus(self, event):
        event.Skip()


class WorldSelectUI(wx.Panel):
    # a frame containing a refresh button for the UI, a sort order for the worlds
    # and a vertical list of `WorldDirectoryUI`s for each directory
    # perhaps also a select directory option
    def __init__(self, parent, open_world_callback):
        super().__init__(parent)
        self.open_world_callback = open_world_callback

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self.header_open_world = wx.Button(
            self, label=lang.get("select_world.open_world_button")
        )
        self.header_open_world.Bind(wx.EVT_BUTTON, self._open_world)
        sizer.Add(self.header_open_world)

        content = ScrollableWorldsUI(self, open_world_callback)
        sizer.Add(content, 1, wx.EXPAND)

    def _open_world(self, evt):
        dir_dialog = wx.DirDialog(
            None,
            lang.get("select_world.open_world_dialogue"),
            "",
            wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST,
        )
        try:
            if dir_dialog.ShowModal() == wx.ID_CANCEL:
                return
            path = dir_dialog.GetPath()
        except Exception:
            wx.LogError(lang.get("select_world.select_directory_failed"))
            return
        finally:
            dir_dialog.Destroy()
        self.open_world_callback(path)


class RecentWorldUI(wx.Panel):
    def __init__(self, parent, open_world_callback):
        super().__init__(parent)
        self._open_world_callback = open_world_callback

        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        text = wx.StaticText(
            self,
            wx.ID_ANY,
            lang.get("select_world.recent_worlds"),
            wx.DefaultPosition,
            wx.DefaultSize,
            0,
        )
        text.SetFont(wx.Font(12, wx.DEFAULT, wx.NORMAL, wx.NORMAL))
        self._sizer.Add(
            text,
            0,
            wx.ALL | wx.ALIGN_CENTER,
            5,
        )

        self._world_list = None
        self.rebuild()

    def rebuild(self, new_world: str = None):
        meta: dict = CONFIG.get("amulet_meta", {})
        recent_worlds: list = meta.setdefault("recent_worlds", [])
        if new_world is not None:
            while new_world in recent_worlds:
                recent_worlds.remove(new_world)
            recent_worlds.insert(0, new_world)
            while len(recent_worlds) > 5:
                recent_worlds.pop(5)
        if self._world_list is not None:
            self._world_list.Destroy()
        self._world_list = WorldList(
            self, recent_worlds, self._open_world_callback, sort=False
        )
        self._sizer.Add(self._world_list, 1, wx.EXPAND, 5)
        self.Layout()
        CONFIG.put("amulet_meta", meta)


class WorldSelectAndRecentUI(wx.Panel):
    def __init__(self, parent, open_world_callback):
        super(WorldSelectAndRecentUI, self).__init__(parent, wx.HORIZONTAL)
        self._open_world_callback = open_world_callback

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        warning_text = wx.StaticText(
            self,
            label=lang.get("select_world.open_world_warning"),
        )
        warning_text.SetFont(wx.Font(20, wx.DEFAULT, wx.NORMAL, wx.NORMAL))
        sizer.Add(warning_text, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 5)
        # bar

        bottom_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(bottom_sizer, 1, wx.EXPAND)

        left_sizer = wx.BoxSizer(wx.VERTICAL)
        bottom_sizer.Add(left_sizer, 1, wx.EXPAND)
        select_world = WorldSelectUI(self, self._update_recent)
        left_sizer.Add(select_world, 1, wx.ALL | wx.EXPAND, 5)

        right_sizer = wx.BoxSizer(wx.VERTICAL)
        bottom_sizer.Add(right_sizer, 1, wx.EXPAND)
        self._recent_worlds = RecentWorldUI(self, self._update_recent)
        right_sizer.Add(self._recent_worlds, 1, wx.EXPAND, 5)

    def _update_recent(self, path):
        self._recent_worlds.rebuild(path)
        self._open_world_callback(path)


@preserve_ui_preferences
class WorldSelectDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, open_world_callback: Callable[[str], None]):
        super().__init__(
            parent,
            title=lang.get("select_world.title"),
            pos=wx.Point(50, 50),
            size=wx.Size(*[int(s * 0.95) for s in parent.GetSize()]),
            style=wx.CAPTION | wx.CLOSE_BOX | wx.MAXIMIZE_BOX
            # | wx.MAXIMIZE
            | wx.SYSTEM_MENU | wx.TAB_TRAVERSAL | wx.CLIP_CHILDREN | wx.RESIZE_BORDER,
        )
        self.Bind(wx.EVT_CLOSE, self._hide_event)

        self._open_world_callback = open_world_callback
        self.world_select = WorldSelectAndRecentUI(self, self._run_callback)

    def _run_callback(self, path):
        self._close()
        self._open_world_callback(path)

    def _hide_event(self, evt):
        self._close()
        evt.Skip()

    def _close(self):
        if self.IsModal():
            self.EndModal(0)
        else:
            self.Close()


def open_level_from_dialog(parent: wx.Window):
    """Show the open world dialog and open the selected world."""
    select_world = WorldSelectDialog(parent, app.open_level)
    select_world.ShowModal()
    select_world.Destroy()



"""
This license applies to this file only.
-- begin license --
MIT License
Copyright (c) 2021 Amulet-Team
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-- end license --
"""

from typing import TYPE_CHECKING
import wx

from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension, OperationReturnType

from amulet_map_editor.programs.edit.api.operations import (
    SimpleOperationPanel,
    OperationSilentAbort,
)

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Clone(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)
        self._add_run_button()
        self.Layout()

    def _operation(
        self, world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
    ) -> OperationReturnType:
        structure = world.extract_structure(selection, dimension)
        self.canvas.paste(structure, structure.dimensions[0])
        raise OperationSilentAbort


export = {
    "name": "Clone",  # the name of the plugin
    "operation": Clone,  # the actual function to call when running the plugin
}



import logging
import warnings
import wx
from typing import Callable, TYPE_CHECKING, Any, Generator, Optional
from types import GeneratorType
from threading import RLock, Thread

from .base_edit_canvas import BaseEditCanvas
from ...edit import EDIT_CONFIG_ID
from ..key_config import (
    DefaultKeys,
    DefaultKeybindGroupId,
    PresetKeybinds,
    KeybindGroup,
)

import time
import traceback

from amulet.api.data_types import OperationReturnType, OperationYieldType, Dimension
from amulet.api.structure import structure_cache
from amulet.api.level import BaseLevel

from amulet_map_editor import CONFIG
from amulet_map_editor import close_level
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from amulet_map_editor.programs.edit.api.ui.goto import show_goto
from amulet_map_editor.programs.edit.api.ui.tool_manager import ToolManagerSizer
from amulet_map_editor.programs.edit.api.operations.errors import (
    OperationError,
    OperationSilentAbort,
    BaseLoudException,
    BaseSilentException,
)
from amulet_map_editor.programs.edit.plugins.operations.stock_plugins.internal_operations import (
    cut,
    copy,
    delete,
)

from amulet_map_editor.programs.edit.api.events import (
    UndoEvent,
    RedoEvent,
    CreateUndoEvent,
    SaveEvent,
    ToolChangeEvent,
    EVT_EDIT_CLOSE,
)
from amulet_map_editor.programs.edit.api.ui.file import FilePanel

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel

log = logging.getLogger(__name__)
OperationType = Callable[[], OperationReturnType]


def show_loading_dialog(
    run: OperationType, title: str, message: str, parent: wx.Window
) -> Any:
    warnings.warn("show_loading_dialog is depreciated.", DeprecationWarning)
    dialog = wx.ProgressDialog(
        title,
        message,
        maximum=10_000,
        parent=parent,
        style=wx.PD_APP_MODAL
        | wx.PD_ELAPSED_TIME
        | wx.PD_REMAINING_TIME
        | wx.PD_AUTO_HIDE,
    )
    dialog.Fit()
    t = time.time()
    try:
        obj = run()
        if isinstance(obj, GeneratorType):
            try:
                while True:
                    progress = next(obj)
                    if isinstance(progress, (list, tuple)):
                        if len(progress) >= 2:
                            message = progress[1]
                        if len(progress) >= 1:
                            progress = progress[0]
                    if isinstance(progress, (int, float)) and isinstance(message, str):
                        dialog.Update(
                            min(9999, max(0, int(progress * 10_000))), message
                        )
                    wx.Yield()
            except StopIteration as e:
                obj = e.value
    except Exception as e:
        dialog.Update(10_000)
        raise e
    time.sleep(max(0.2 - time.time() + t, 0))
    dialog.Update(10_000)
    return obj


class OperationThread(Thread):
    # The operation to run
    _operation: OperationType

    # Should the operation be stopped. Set externally
    stop: bool
    # The starting message for the progress dialog
    message: str
    # The operation progress (from 0-1)
    progress: float
    # The return value from the operation
    out: Any
    # The error raised if any
    error: Optional[BaseException]

    def __init__(self, operation: OperationType, message: str):
        super().__init__()
        self._operation = operation
        self.stop = False
        self.message = message
        self.progress = 0.0
        self.out = None
        self.error = None

    def run(self) -> None:
        t = time.time()
        try:
            obj = self._operation()
            if isinstance(obj, GeneratorType):
                try:
                    while True:
                        if self.stop:
                            raise OperationSilentAbort
                        progress = next(obj)
                        if isinstance(progress, (list, tuple)):
                            if len(progress) >= 2:
                                self.message = progress[1]
                            if len(progress) >= 1:
                                self.progress = progress[0]
                        elif isinstance(progress, (int, float)):
                            self.progress = progress
                except StopIteration as e:
                    self.out = e.value
        except BaseException as e:
            self.error = e
        time.sleep(max(0.2 - time.time() + t, 0))


class EditCanvas(BaseEditCanvas):
    def __init__(self, parent: wx.Window, world: "BaseLevel"):
        super().__init__(parent, world)
        self._file_panel: Optional[FilePanel] = None
        self._tool_sizer: Optional[ToolManagerSizer] = None
        self.buttons.register_actions(self.key_binds)

        self._canvas_sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._canvas_sizer)

        # Tracks if an operation has been started and not finished.
        self._operation_running = False
        # This lock stops two threads from editing the world simultaneously
        # call run_operation to acquire it.
        self._edit_lock = RLock()

    def _init_opengl(self):
        super()._init_opengl()
        self._file_panel = FilePanel(self)
        self._canvas_sizer.Add(self._file_panel, 0, wx.EXPAND, 0)
        self._tool_sizer = ToolManagerSizer(self)
        self._canvas_sizer.Add(self._tool_sizer, 1, wx.EXPAND, 0)

    def bind_events(self):
        """Set up all events required to run.
        Note this will also bind subclass events."""
        self._tool_sizer.bind_events()
        # binding the tool events first will run them last so they can't accidentally block UI events.
        super().bind_events()
        self._file_panel.bind_events()
        self.Bind(EVT_EDIT_CLOSE, self._on_close)

    def enable(self):
        super().enable()
        self._tool_sizer.enable()

    def disable(self):
        super().disable()
        self._tool_sizer.disable()

    def _on_close(self, _):
        close_level(self.world.level_path)

    @property
    def tools(self):
        return self._tool_sizer.tools

    @property
    def key_binds(self) -> KeybindGroup:
        config_ = CONFIG.get(EDIT_CONFIG_ID, {})
        user_keybinds = config_.get("user_keybinds", {})
        group = config_.get("keybind_group", DefaultKeybindGroupId)
        if group in user_keybinds:
            return user_keybinds[group]
        elif group in PresetKeybinds:
            return PresetKeybinds[group]
        else:
            return DefaultKeys

    def _deselect(self):
        # TODO: Re-implement this
        self._tool_sizer.enable_default_tool()

    def run_operation(
        self,
        operation: OperationType,
        title="Amulet",
        msg="Running Operation",
        throw_exceptions=False,
    ) -> Any:
        try:
            out = self._run_operation(operation, title, msg, True)
        except Exception as e:
            if throw_exceptions:
                raise e
        else:
            # If there were no errors create an undo point
            def create_undo():
                yield 0, "Creating Undo Point"
                yield from self.create_undo_point_iter()

            self._run_operation(create_undo, title, msg, False)

            return out

    def _run_operation(
        self,
        operation: OperationType,
        title: str,
        msg: str,
        cancelable: bool,
    ) -> Any:
        with self._edit_lock:
            if self._operation_running:
                raise Exception(
                    "run_operation cannot be called from within itself. "
                    "This function has already been called by parent code so you cannot run it again"
                )
            self._operation_running = True

            self.renderer.disable_threads()

            style = (
                wx.PD_APP_MODAL
                | wx.PD_ELAPSED_TIME
                | wx.PD_REMAINING_TIME
                | wx.PD_AUTO_HIDE
                | (wx.PD_CAN_ABORT * cancelable)
            )
            dialog = wx.ProgressDialog(
                title,
                msg,
                maximum=10_000,
                parent=self,
                style=style,
            )
            dialog.Fit()

            # Set up a thread to run the actual operation
            op = OperationThread(operation, msg)
            # run the operation
            op.start()
            while op.is_alive():
                op.join(0.1)
                dialog.Update(max(0, min(int(op.progress * 10_000), 9999)), op.message)
                wx.Yield()
                if dialog.WasCancelled():
                    op.stop = True

            dialog.Destroy()
            wx.Yield()

            if op.error is not None:
                # If there is any kind of error restore the last undo point
                self.world.restore_last_undo_point()

                if isinstance(op.error, BaseLoudException):
                    msg = str(op.error)
                    if isinstance(op.error, OperationError):
                        msg = f"Error running operation: {msg}"
                    log.info(msg)
                    wx.MessageDialog(self, msg, style=wx.OK).ShowModal()
                elif isinstance(op.error, BaseSilentException):
                    pass
                elif isinstance(op.error, BaseException):
                    log.error(traceback.format_exc())
                    dialog = TracebackDialog(
                        self,
                        "Exception while running operation",
                        str(op.error),
                        traceback.format_exc(),
                    )
                    dialog.ShowModal()
                    dialog.Destroy()
                    self.world.restore_last_undo_point()

            self.renderer.enable_threads()
            self.renderer.render_world.rebuild_changed()
            self._operation_running = False
            if op.error is not None:
                raise op.error
            return op.out

    def create_undo_point(self, world=True, non_world=True):
        self.world.create_undo_point(world, non_world)
        wx.PostEvent(self, CreateUndoEvent())

    def create_undo_point_iter(
        self, world=True, non_world=True
    ) -> Generator[float, None, bool]:
        result = yield from self.world.create_undo_point_iter(world, non_world)
        wx.PostEvent(self, CreateUndoEvent())
        return result

    def undo(self):
        self.world.undo()
        self.renderer.render_world.rebuild_changed()
        wx.PostEvent(self, UndoEvent())

    def redo(self):
        self.world.redo()
        self.renderer.render_world.rebuild_changed()
        wx.PostEvent(self, RedoEvent())

    def cut(self):
        self.run_operation(
            lambda: cut(self.world, self.dimension, self.selection.selection_group)
        )

    def copy(self):
        self.run_operation(
            lambda: copy(self.world, self.dimension, self.selection.selection_group)
        )

    def paste(self, structure: BaseLevel, dimension: Dimension):
        assert isinstance(
            structure, BaseLevel
        ), "Structure given is not a subclass of BaseLevel."
        assert (
            dimension in structure.dimensions
        ), "The requested dimension does not exist for this object."
        wx.PostEvent(
            self,
            ToolChangeEvent(
                tool="Paste", state={"structure": structure, "dimension": dimension}
            ),
        )

    def paste_from_cache(self):
        if structure_cache:
            self.paste(*structure_cache.get_structure())
        else:
            wx.MessageBox("A structure needs to be copied before one can be pasted.")

    def delete(self):
        self.run_operation(
            lambda: delete(self.world, self.dimension, self.selection.selection_group)
        )

    def goto(self):
        location = show_goto(self, *self.camera.location)
        if location:
            self.camera.location = location

    def select_all(self):
        all_chunk_coords = tuple(self.world.all_chunk_coords(self.dimension))
        if all_chunk_coords:
            min_x, min_z = max_x, max_z = all_chunk_coords[0]
            for x, z in all_chunk_coords:
                if x < min_x:
                    min_x = x
                elif x > max_x:
                    max_x = x
                if z < min_z:
                    min_z = z
                elif z > max_z:
                    max_z = z

            self.selection.selection_corners = [
                (
                    (
                        min_x * self.world.sub_chunk_size,
                        self.world.bounds(self.dimension).min[1],
                        min_z * self.world.sub_chunk_size,
                    ),
                    (
                        (max_x + 1) * self.world.sub_chunk_size,
                        self.world.bounds(self.dimension).max[1],
                        (max_z + 1) * self.world.sub_chunk_size,
                    ),
                )
            ]

        else:
            self.selection.selection_corners = []

    def save(self):
        def save():
            yield 0, "Running Pre-Save Operations."
            pre_save_op = self.world.pre_save_operation()
            try:
                while True:
                    yield next(pre_save_op)
            except StopIteration as e:
                if e.value:
                    yield from self.create_undo_point_iter()
                else:
                    self.world.restore_last_undo_point()

            yield 0, "Saving Chunks."
            for chunk_index, chunk_count in self.world.save_iter():
                yield chunk_index / chunk_count

        self._run_operation(save, "Saving world.", "Please wait.", False)
        wx.PostEvent(self, SaveEvent())



from typing import TYPE_CHECKING, Optional, Generator
import webbrowser
import logging
from threading import Thread
import traceback

import wx

from amulet.api.data_types import OperationYieldType

EDIT_CONFIG_ID = "amulet_edit"

from amulet_map_editor import lang
from amulet_map_editor.api.framework.programs import BaseProgram
from amulet_map_editor.api.datatypes import MenuData
from amulet_map_editor.api.wx.util.key_config import KeyConfigDialog
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from amulet_map_editor.api.wx.ui.simple import SimpleDialog
from amulet_map_editor.programs.edit.api.canvas.edit_canvas import EditCanvas
from amulet_map_editor.programs.edit.api.key_config import (
    DefaultKeybindGroupId,
    PresetKeybinds,
    KeybindKeys,
)
from amulet_map_editor.api import config

if TYPE_CHECKING:
    from amulet.api.level import World

log = logging.getLogger(__name__)


class EditExtension(wx.Panel, BaseProgram):

    # UI elements
    _sizer: wx.BoxSizer
    # these only exists on setup. Once setup is finished they will be None
    _temp_msg: Optional[wx.StaticText]
    _temp_loading_bar: Optional[wx.Gauge]

    _world: "World"
    _canvas: Optional[EditCanvas]

    # setup is run in a different thread to avoid blocking the UI
    _setup_thread: Optional[Thread]

    def __init__(self, parent, world: "World"):
        wx.Panel.__init__(self, parent)
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetBackgroundColour(
            tuple(int(v * 255) for v in EditCanvas.background_colour)
        )
        self.SetSizer(self._sizer)
        self._world = world
        self._canvas = None
        self._setup_thread = None

        self._sizer.AddStretchSpacer(1)
        self._temp_msg = wx.StaticText(
            self, label=lang.get("program_3d_edit.canvas.please_wait")
        )
        self._temp_msg.SetFont(wx.Font(40, wx.DECORATIVE, wx.NORMAL, wx.NORMAL))
        self._sizer.Add(self._temp_msg, 0, flag=wx.ALIGN_CENTER_HORIZONTAL)
        self._temp_loading_bar = wx.Gauge(self, range=10000)
        self._sizer.Add(self._temp_loading_bar, 0, flag=wx.EXPAND)
        self._sizer.AddStretchSpacer(1)

    def enable(self):
        if self._canvas is None:
            self._canvas = EditCanvas(self, self._world)
            self._canvas.Hide()
            self._setup_thread = Thread(target=self._thread_setup)
            self._setup_thread.start()
        else:
            self._canvas.enable()

    def _update_loading(self, it: Generator[OperationYieldType, None, None]):
        for arg in it:
            if isinstance(arg, (int, float)):
                self._temp_loading_bar.SetValue(int(min(arg, 1) * 10000))
            elif (
                isinstance(arg, tuple)
                and isinstance(arg[0], (int, float))
                and isinstance(arg[1], str)
            ):
                self._temp_loading_bar.SetValue(int(min(arg[0], 1) * 10000))
                self._temp_msg.SetLabel(arg[1])
            self.Layout()

    def _display_error(self, msg, tb):
        dialog = TracebackDialog(
            self,
            "Exception while setting up canvas",
            msg,
            tb,
        )
        dialog.ShowModal()
        dialog.Destroy()
        self.Destroy()

    def _thread_setup(self):
        """
        Setup and enable all the UI elements.
        This can take a while to run so should be done in a new thread.
        Everything in here must be thread safe.
        """
        try:
            self._update_loading(self._canvas.thread_setup())
        except Exception as e:
            wx.CallAfter(self._display_error, str(e), traceback.format_exc())
            raise e
        else:
            wx.CallAfter(self._post_thread_setup)

    def _post_thread_setup(self):
        """
        Run any setup that is not thread safe.
        """
        try:
            self._update_loading(self._canvas.post_thread_setup())
            edit_config: dict = config.get(EDIT_CONFIG_ID, {})
            self._canvas.camera.perspective_fov = edit_config.get("options", {}).get(
                "fov", 70.0
            )
            if self._canvas.camera.perspective_fov > 180:
                self._canvas.camera.perspective_fov = 70.0
            self._canvas.renderer.render_distance = edit_config.get("options", {}).get(
                "render_distance", 5
            )
            self._canvas.camera.rotate_speed = edit_config.get("options", {}).get(
                "camera_sensitivity", 2.0
            )

            self._temp_msg = None
            self._temp_loading_bar = None
            self._sizer.Clear(True)
            self._sizer.Add(self._canvas, 1, wx.EXPAND)
            self._canvas.Show()
            self._canvas._set_size()
            self.Layout()
            wx.CallAfter(
                self._canvas.enable
            )  # This must be called after the show handler is run
            self._setup_thread = None
        except Exception as e:
            wx.CallAfter(self._display_error, str(e), traceback.format_exc())
            raise e

    def can_disable(self) -> bool:
        return self._setup_thread is None

    def disable(self):
        if self._canvas is not None:
            self._canvas.disable()

    def close(self):
        """Fully close the UI. Called when destroying the UI."""
        if self._canvas is not None:
            self._canvas.close()

    def can_close(self) -> bool:
        """
        Check if it is safe to close the UI.
        :return: True if the program can be closed, False otherwise
        """
        if self._setup_thread is not None:
            return False
        elif self._canvas is None:
            return True
        elif self._canvas.is_closeable():
            return self._check_close_world()
        else:
            log.info(
                f"The canvas in edit for world {self._world.level_wrapper.level_name} was not closeable for some reason."
            )
            return False

    def _check_close_world(self) -> bool:
        """
        Check if it is safe to close the world and prompt the user if it is not.
        :return: True if the world can be closed, False otherwise
        """
        unsaved_changes = self._world.history_manager.unsaved_changes
        if unsaved_changes:
            msg = wx.MessageDialog(
                self,
                f"""There {
                'is' if unsaved_changes == 1 else 'are'
                } {unsaved_changes} unsaved change{
                's' if unsaved_changes >= 2 else ''
                } in {
                self._world.level_wrapper.level_name
                }. Would you like to save?""",
                style=wx.YES_NO | wx.CANCEL | wx.CANCEL_DEFAULT,
            )
            response = msg.ShowModal()
            if response == wx.ID_YES:
                self._canvas.save()
                return True
            elif response == wx.ID_NO:
                return True
            elif response == wx.ID_CANCEL:
                log.info(
                    f"""Aborting closing world {
                    self._world.level_wrapper.level_name
                    } because the user pressed cancel."""
                )
                return False
        return True

    def menu(self, menu: MenuData) -> MenuData:
        menu.setdefault(lang.get("menu_bar.file.menu_name"), {}).setdefault(
            "system", {}
        ).setdefault(
            f"{lang.get('program_3d_edit.menu_bar.file.save')}\tCtrl+s",
            lambda evt: self._canvas.save(),
        )
        # menu.setdefault(lang.get('menu_bar.file.menu_name'), {}).setdefault('system', {}).setdefault('Save As', lambda evt: self.GetGrandParent().close_world(self.world.world_path))

        menu.setdefault(
            lang.get("program_3d_edit.menu_bar.edit.menu_name"), {}
        ).setdefault("history", {}).update(
            {
                f"{lang.get('program_3d_edit.menu_bar.edit.undo')}\tCtrl+z": lambda evt: self._canvas.undo(),
                f"{lang.get('program_3d_edit.menu_bar.edit.redo')}\tCtrl+y": lambda evt: self._canvas.redo(),
            }
        )

        menu.setdefault(
            lang.get("program_3d_edit.menu_bar.edit.menu_name"), {}
        ).setdefault("operation", {}).update(
            {
                f"{lang.get('program_3d_edit.menu_bar.edit.cut')}\tCtrl+x": lambda evt: self._canvas.cut(),
                f"{lang.get('program_3d_edit.menu_bar.edit.copy')}\tCtrl+c": lambda evt: self._canvas.copy(),
                f"{lang.get('program_3d_edit.menu_bar.edit.paste')}\tCtrl+v": lambda evt: self._canvas.paste_from_cache(),
                f"{lang.get('program_3d_edit.menu_bar.edit.delete')}\tDelete": lambda evt: self._canvas.delete(),
            }
        )

        menu.setdefault(
            lang.get("program_3d_edit.menu_bar.edit.menu_name"), {}
        ).setdefault("shortcut", {}).update(
            {
                f"{lang.get('program_3d_edit.menu_bar.edit.goto')}\tCtrl+g": lambda evt: self._canvas.goto(),
                f"{lang.get('program_3d_edit.menu_bar.edit.select_all')}\tCtrl+A": lambda evt: self._canvas.select_all(),
            }
        )

        menu.setdefault(lang.get("menu_bar.options.menu_name"), {}).setdefault(
            "options", {}
        ).setdefault(
            lang.get("program_3d_edit.menu_bar.options.controls"),
            lambda evt: self._edit_controls(),
        )
        menu.setdefault(lang.get("menu_bar.options.menu_name"), {}).setdefault(
            "options", {}
        ).setdefault(
            lang.get("program_3d_edit.menu_bar.options.options"),
            lambda evt: self._edit_options(),
        )
        menu.setdefault(lang.get("menu_bar.help.menu_name"), {}).setdefault(
            "help", {}
        ).setdefault(
            lang.get("program_3d_edit.menu_bar.help.user_guide"),
            lambda evt: self._help_controls(),
        )
        return menu

    def _edit_controls(self):
        edit_config = config.get(EDIT_CONFIG_ID, {})
        keybind_id = edit_config.get("keybind_group", DefaultKeybindGroupId)
        user_keybinds = edit_config.get("user_keybinds", {})
        key_config = KeyConfigDialog(
            self, keybind_id, KeybindKeys, PresetKeybinds, user_keybinds
        )
        if key_config.ShowModal() == wx.ID_OK:
            user_keybinds, keybind_id, keybinds = key_config.options
            edit_config["user_keybinds"] = user_keybinds
            edit_config["keybind_group"] = keybind_id
            config.put(EDIT_CONFIG_ID, edit_config)
            self._canvas.buttons.clear_registered_actions()
            self._canvas.buttons.register_actions(keybinds)

    def _edit_options(self):
        if self._canvas is not None:
            fov = self._canvas.camera.perspective_fov
            render_distance = self._canvas.renderer.render_distance
            camera_sensitivity = self._canvas.camera.rotate_speed
            dialog = SimpleDialog(self, "Options")

            sizer = wx.FlexGridSizer(3, 2, 0, 0)
            dialog.sizer.Add(sizer, flag=wx.ALL, border=5)
            fov_ui = wx.SpinCtrlDouble(dialog, min=0, max=180, initial=fov)

            def set_fov(evt):
                self._canvas.camera.perspective_fov = fov_ui.GetValue()

            fov_ui.Bind(wx.EVT_SPINCTRLDOUBLE, set_fov)
            sizer.Add(
                wx.StaticText(dialog, label="Field of View"),
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )
            sizer.Add(
                fov_ui,
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )

            render_distance_ui = wx.SpinCtrl(
                dialog, min=0, max=500, initial=render_distance
            )

            def set_render_distance(evt):
                self._canvas.renderer.render_distance = render_distance_ui.GetValue()

            render_distance_ui.Bind(wx.EVT_SPINCTRL, set_render_distance)
            sizer.Add(
                wx.StaticText(dialog, label="Render Distance"),
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )
            sizer.Add(
                render_distance_ui,
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )

            camera_sensitivity_ui = wx.SpinCtrlDouble(
                dialog, min=0, max=10, initial=camera_sensitivity
            )

            def set_camera_sensitivity(evt):
                self._canvas.camera.rotate_speed = camera_sensitivity_ui.GetValue()

            camera_sensitivity_ui.Bind(wx.EVT_SPINCTRLDOUBLE, set_camera_sensitivity)
            sizer.Add(
                wx.StaticText(dialog, label="Camera Sensitivity"),
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )
            sizer.Add(
                camera_sensitivity_ui,
                flag=wx.LEFT | wx.TOP | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                border=5,
            )

            dialog.Fit()

            response = dialog.ShowModal()
            if response == wx.ID_OK:
                edit_config: dict = config.get(EDIT_CONFIG_ID, {})
                edit_config.setdefault("options", {})
                edit_config["options"]["fov"] = fov_ui.GetValue()
                edit_config["options"][
                    "render_distance"
                ] = render_distance_ui.GetValue()
                edit_config["options"][
                    "camera_sensitivity"
                ] = camera_sensitivity_ui.GetValue()
                config.put(EDIT_CONFIG_ID, edit_config)
            elif response == wx.ID_CANCEL:
                self._canvas.camera.perspective_fov = fov
                self._canvas.renderer.render_distance = render_distance
                self._canvas.camera.rotate_speed = camera_sensitivity

    @staticmethod
    def _help_controls():
        webbrowser.open(
            "https://github.com/Amulet-Team/Amulet-Map-Editor/blob/master/amulet_map_editor/programs/edit/readme.md"
        )



import numpy
from OpenGL.GL import (
    GL_TRIANGLES,
    GL_DYNAMIC_DRAW,
    glCullFace,
    GL_FRONT,
    GL_BACK,
    glDisable,
    glEnable,
    GL_DEPTH_TEST,
    GL_LINE_STRIP,
    glGetBooleanv,
    glGetIntegerv,
    GL_CULL_FACE_MODE,
)
import itertools
from typing import Tuple, Optional, Union

from amulet.api.selection import SelectionBox, SelectionGroup
from amulet_map_editor.api.opengl.mesh.tri_mesh import TriMesh
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePack,
    OpenGLResourcePackManagerStatic,
)
from amulet.api.data_types import BlockCoordinatesAny, PointCoordinatesAny
from amulet_map_editor.api.opengl.matrix import displacement_matrix
from amulet_map_editor.api.opengl.data_types import RGBColour
from .colours import colours


class RenderSelection(TriMesh, OpenGLResourcePackManagerStatic):
    """A drawable selection box"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        TriMesh.__init__(self, context_identifier, resource_pack)
        self._points: numpy.ndarray = numpy.zeros(
            (2, 3), dtype=numpy.int64
        )  # The points set using point1 and point2
        self._bounds: Optional[numpy.ndarray] = None  # The min and max locations
        self.transformation_matrix = numpy.eye(4, dtype=numpy.float64)
        self._needs_rebuild = True
        self._volume = 1

        self._init_verts()
        self.draw_count = 36
        self._draw_mode = GL_TRIANGLES

    @property
    def box_tint(self) -> RGBColour:
        """The tint colour to apply to the white texture."""
        return colours.get("box_normal", (1, 1, 1))

    def _init_verts(self):
        """Initialise the vertex values"""
        self.verts = numpy.zeros((6 * 2 * 3, self._vert_len), dtype=numpy.float32)
        self.verts[:36, 5:9] = self.resource_pack.texture_bounds(
            self.resource_pack.get_texture_path("amulet", "amulet_ui/selection")
        )
        self.verts[:, 9:12] = self.box_tint

    def __contains__(
        self, position: Union[BlockCoordinatesAny, PointCoordinatesAny]
    ) -> bool:
        """
        Is the block position inside the selection box cuboid.
        :param position: (x, y, z)
        :return: True if the position is inside the box otherwise False
        """
        point = numpy.array(position)
        return numpy.all(self.min <= point) and numpy.all(point < self.max)

    def _offset_points(self) -> numpy.ndarray:
        points = self._points.copy()
        points[0] -= self.point1 > self.point2
        points[1] -= self.point1 <= self.point2
        return points

    def _from_offset_points(self, offset_points: numpy.ndarray):
        points = offset_points.copy()
        points[0] += offset_points[0] > offset_points[1]
        points[1] += offset_points[0] <= offset_points[1]
        self._points[:] = points

    @property
    def vertex_usage(self):
        return GL_DYNAMIC_DRAW

    @property
    def draw_mode(self):
        return self._draw_mode

    def _mark_recreate(self):
        self._bounds = None
        self._needs_rebuild = True

    @property
    def points(self) -> numpy.ndarray:
        return self._points.copy()

    @points.setter
    def points(self, points: numpy.ndarray):
        if not isinstance(points, numpy.ndarray) and points.shape == (2, 3):
            raise TypeError("points must be a numpy array of size 2x3.")
        self.point1, self.point2 = points

    @property
    def point1(self) -> numpy.ndarray:
        return self.points[0]

    @point1.setter
    def point1(self, val: PointCoordinatesAny):
        if not numpy.array_equal(self._points[0], val):
            self._points[0] = numpy.floor(val)
            self._mark_recreate()

    @property
    def point2(self) -> numpy.ndarray:
        return self.points[1]

    @point2.setter
    def point2(self, val: PointCoordinatesAny):
        if not numpy.array_equal(self._points[1], val):
            self._points[1] = numpy.floor(val)
            self._mark_recreate()

    @property
    def bounds(self) -> numpy.ndarray:
        """The array storing min and max locations"""
        if self._bounds is None:
            self._bounds = numpy.sort(self._points, 0)
        return self._bounds

    @property
    def selection_box(self) -> SelectionBox:
        return SelectionBox(self.point1, self.point2)

    @selection_box.setter
    def selection_box(self, selection_box: SelectionBox):
        if not isinstance(selection_box, SelectionBox):
            raise TypeError("selection_box must be a SelectionBox.")
        self.point1 = selection_box.point_1
        self.point2 = selection_box.point_2

    @property
    def selection_group(self) -> SelectionGroup:
        return SelectionGroup(self.selection_box)

    @property
    def min(self) -> numpy.ndarray:
        return self.bounds[0]

    @property
    def max(self) -> numpy.ndarray:
        return self.bounds[1]

    def _create_box(
        self, box_min: PointCoordinatesAny, box_max: PointCoordinatesAny
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        return self._create_box_faces(
            box_min, box_max, True, True, True, True, True, True
        )

    @staticmethod
    def _create_box_faces(
        box_min: PointCoordinatesAny,
        box_max: PointCoordinatesAny,
        up: bool = False,
        down: bool = False,
        north: bool = False,
        south: bool = False,
        east: bool = False,
        west: bool = False,
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        box = numpy.sort([box_min, box_max], 0)
        _face_count = sum([up, down, north, south, east, west])
        if not _face_count:
            raise ValueError("At least one of the faces must be enabled.")
        _box_coordinates = numpy.array(list(itertools.product(*box.T.tolist())))
        _cube_face_lut = numpy.array(
            [  # This maps to the vertices used (defined in cube_vert_lut)
                0,
                4,
                5,
                1,  # down
            ]
            * down
            + [0, 1, 3, 2] * west  # west
            + [4, 0, 2, 6] * north  # north
            + [5, 4, 6, 7] * east  # east
            + [1, 5, 7, 3] * south  # south
            + [3, 7, 6, 2] * up  # up
        )
        box = box.ravel()
        _texture_index = numpy.array(
            [0, 2, 3, 5] * down  # down
            + [2, 1, 5, 4] * west  # west
            + [3, 1, 0, 4] * north  # north
            + [5, 1, 2, 4] * east  # east
            + [0, 1, 3, 4] * south  # south
            + [0, 5, 3, 2] * up,  # up
            numpy.uint32,
        )
        _uv_slice = numpy.array(
            [0, 1, 2, 1, 2, 3, 0, 3] * _face_count, dtype=numpy.uint32
        ).reshape((_face_count, 8)) + numpy.arange(0, _face_count * 4, 4).reshape(
            (_face_count, 1)
        )

        _tri_face = numpy.array([0, 1, 2, 2, 3, 0] * _face_count, numpy.uint32).reshape(
            (_face_count, 6)
        ) + numpy.arange(0, _face_count * 4, 4).reshape((_face_count, 1))
        return (
            _box_coordinates[_cube_face_lut[_tri_face]].reshape((-1, 3)),
            box[_texture_index[_uv_slice]]
            .reshape(-1, 2)[_tri_face, :]
            .reshape((-1, 2)),
        )

    def _create_geometry_(self):
        self.verts[:36, :3], self.verts[:36, 3:5] = self._create_box(
            self.min % 16 - 0.005, self.min % 16 + self.max - self.min + 0.005
        )
        self.verts[:36, 3:5] /= 16

    def _create_geometry(self):
        self._setup()
        self._create_geometry_()

        self.transformation_matrix = displacement_matrix(*self.min - self.min % 16)

        self.change_verts()
        self._volume = numpy.product(self.max - self.min)
        self._needs_rebuild = False

    def draw(
        self, camera_matrix: numpy.ndarray, camera_position: PointCoordinatesAny = None
    ):
        """
        Draw the selection box
        :param camera_matrix: 4x4 transformation matrix for the camera
        :param camera_position: The position of the camera. Used to flip draw direction if camera inside box.
        :return:
        """
        self._setup()
        if self._needs_rebuild:
            self._create_geometry()
        self._draw_mode = GL_TRIANGLES

        transformation_matrix = numpy.matmul(camera_matrix, self.transformation_matrix)

        depth_state = glGetBooleanv(GL_DEPTH_TEST)
        cull_state = glGetIntegerv(GL_CULL_FACE_MODE)

        if camera_position is not None and camera_position in self:
            if cull_state == GL_BACK:
                glCullFace(GL_FRONT)
            elif cull_state == GL_FRONT:
                glCullFace(GL_BACK)

        self.draw_start = 0
        self.draw_count = 36
        super()._draw(transformation_matrix)

        # draw the lines around the boxes
        if depth_state:
            glDisable(GL_DEPTH_TEST)
        self._draw_mode = GL_LINE_STRIP
        super()._draw(transformation_matrix)
        if depth_state:
            glEnable(GL_DEPTH_TEST)

        glCullFace(cull_state)



from typing import Tuple
import numpy


TransformationMatrix = numpy.ndarray  # 4x4 matrix
CameraLocationType = Tuple[float, float, float]
CameraRotationType = Tuple[float, float]

RGBColour = Tuple[float, float, float]



import logging
import wx
import traceback
from OpenGL.GL import (
    glClearColor,
    glViewport,
)
import os
from typing import Optional, Generator
import weakref
import sys
import time

from minecraft_model_reader.api.resource_pack.java.download_resources import (
    get_java_vanilla_latest_iter,
    get_java_vanilla_fix,
)
from minecraft_model_reader.api.resource_pack.java import (
    JavaResourcePack,
)
from minecraft_model_reader.api.resource_pack.bedrock.download_resources import (
    get_bedrock_vanilla_latest_iter,
    get_bedrock_vanilla_fix,
)
from minecraft_model_reader.api.resource_pack.bedrock import BedrockResourcePack
from minecraft_model_reader.api.resource_pack import (
    load_resource_pack,
    load_resource_pack_manager,
)
from amulet.api.player import LOCAL_PLAYER

from amulet.api.data_types import OperationYieldType, Dimension

from amulet_map_editor import experimental_bedrock_resources
from amulet_map_editor.api.opengl.canvas import EventCanvas
from amulet_map_editor.api.opengl.resource_pack.resource_pack import OpenGLResourcePack
from amulet_map_editor.programs.edit.api.selection import (
    SelectionManager,
    SelectionHistoryManager,
)
from amulet_map_editor import lang
import amulet_map_editor.programs.edit as amulet_edit

from amulet_map_editor.api.opengl.camera import ControllableCamera
from amulet_map_editor.api.wx.util.button_input import ButtonInput
from amulet_map_editor.api.wx.util.mouse_movement import MouseMovement
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from ..renderer import Renderer

from amulet.api.level import BaseLevel

log = logging.getLogger(__name__)


class BaseEditCanvas(EventCanvas):
    """Adds base logic for drawing everything related to the edit program to the canvas.
    All the user interaction code is implemented in ControllableEditCanvas to make them easier to read."""

    background_colour = (0.61, 0.70, 0.85)

    def __init__(self, parent: wx.Window, world: BaseLevel):
        super().__init__(parent)
        self._world = weakref.ref(world)

        self._selection: Optional[SelectionHistoryManager] = SelectionHistoryManager(
            SelectionManager(self)
        )
        self.world.history_manager.register(self._selection, False)

        self._camera: ControllableCamera = ControllableCamera(self)

        self._camera.move_speed = 2.0
        self._camera.rotate_speed = 2.0

        self._buttons: ButtonInput = ButtonInput(self)
        self._mouse: MouseMovement = MouseMovement(self)
        self._mouse.set_middle()

        # create the resource packs location
        try:
            os.makedirs("resource_packs", exist_ok=True)
            if not os.path.isfile("resource_packs/readme.txt"):
                with open("resource_packs/readme.txt", "w") as f:
                    f.write("Put the Java resource pack you want loaded in here.")
        except PermissionError as e:
            raise PermissionError(
                "Amulet is not able to write to the install directory. Try moving Amulet to somewhere else on your computer."
            ) from e

        self._renderer: Optional[Renderer] = None
        self._opengl_resource_pack = None

    def thread_setup(self) -> Generator[OperationYieldType, None, None]:
        """
        Set up objects that take a while to set up.
        All code in here must be thread safe and not touch the OpenGL state.
        """
        packs = []
        user_packs = [
            load_resource_pack(os.path.join("resource_packs", rp))
            for rp in os.listdir("resource_packs")
            if os.path.isdir(os.path.join("resource_packs", rp))
        ]
        if (
            self.world.level_wrapper.platform == "bedrock"
            and experimental_bedrock_resources
        ):
            packs.append(
                load_resource_pack(
                    os.path.join(
                        os.path.dirname(amulet_edit.__file__),
                        "amulet_resource_pack",
                        "bedrock",
                    )
                )
            )
            yield 0.1, lang.get(
                "program_3d_edit.canvas.downloading_bedrock_vanilla_resource_pack"
            )
            gen = get_bedrock_vanilla_latest_iter()
            try:
                while True:
                    yield next(gen) * 0.4 + 0.1
            except StopIteration as e:
                packs.append(e.value)
            yield 0.5, lang.get("program_3d_edit.canvas.loading_resource_packs")

            packs += [
                pack for pack in user_packs if isinstance(pack, BedrockResourcePack)
            ]
            packs.append(get_bedrock_vanilla_fix())

            translator = self.world.translation_manager.get_version(
                "bedrock", (999, 0, 0)
            )
        else:
            packs.append(
                load_resource_pack(
                    os.path.join(
                        os.path.dirname(amulet_edit.__file__),
                        "amulet_resource_pack",
                        "java",
                    )
                )
            )
            yield 0.1, lang.get(
                "program_3d_edit.canvas.downloading_java_vanilla_resource_pack"
            )
            gen = get_java_vanilla_latest_iter()
            try:
                while True:
                    yield next(gen) * 0.4 + 0.1
            except StopIteration as e:
                packs.append(e.value)
            except Exception as e:
                if sys.platform == "darwin" and "CERTIFICATE_VERIFY_FAILED" in str(e):
                    msg = lang.get(
                        "program_3d_edit.canvas.java_rp_failed_mac_certificates"
                    )
                else:
                    msg = lang.get("program_3d_edit.canvas.java_rp_failed_default")
                log.error(
                    msg,
                    exc_info=True,
                )
                wait = True

                def show_error():
                    nonlocal wait
                    try:
                        dialog = TracebackDialog(
                            self,
                            lang.get("program_3d_edit.canvas.java_rp_failed"),
                            f"{msg}\n{e}",
                            traceback.format_exc(),
                        )
                        dialog.ShowModal()
                        dialog.Destroy()
                    finally:
                        wait = False

                wx.CallAfter(show_error)
                while wait:
                    time.sleep(0.1)

            yield 0.5, lang.get("program_3d_edit.canvas.loading_resource_packs")
            packs += [pack for pack in user_packs if isinstance(pack, JavaResourcePack)]
            packs.append(get_java_vanilla_fix())

            translator = self.world.translation_manager.get_version("java", (999, 0, 0))

        resource_pack = load_resource_pack_manager(packs, load=False)
        for i in resource_pack.reload():
            yield i / 4 + 0.5

        self._opengl_resource_pack = OpenGLResourcePack(resource_pack, translator)

        yield 0.75, lang.get("program_3d_edit.canvas.creating_texture_atlas")
        for i in self._opengl_resource_pack.setup():
            yield i / 4 + 0.75

    def post_thread_setup(self) -> Generator[OperationYieldType, None, None]:
        """
        Any logic that needs to be run after everything has been set up.
        Code here will be run from the main thread.
        The canvas has still not been shown here so do not touch the OpenGL state. Do this in _init_opengl.
        """
        yield 1.0, lang.get("program_3d_edit.canvas.setting_up_renderer")
        self._renderer: Optional[Renderer] = Renderer(
            self,
            self.world,
            self.context_identifier,
            self._opengl_resource_pack,
        )

    def _init_opengl(self):
        super()._init_opengl()
        glClearColor(*self.background_colour, 1.0)

        try:
            player = self.world.get_player(LOCAL_PLAYER)
            location, rotation = player.location, player.rotation
            self.dimension = player.dimension
        except:
            location, rotation = (0.0, 100.0, 0.0), (45.0, 45.0)

        self._camera.location_rotation = location, rotation
        self._renderer.move_camera(location, rotation)

    def bind_events(self):
        """Set up all events required to run.
        Note this will also bind subclass events."""
        self.Bind(wx.EVT_SIZE, self._on_size)
        self.selection.bind_events()
        self.buttons.bind_events()
        self.mouse.bind_events()
        self.renderer.bind_events()

    def enable(self):
        """Enable the canvas and start it working."""
        self.SetCurrent(self._context)
        self.renderer.enable()
        self.buttons.enable()

    def disable(self):
        """Disable the canvas and unload all geometry."""
        self.renderer.disable()
        self.buttons.disable()

    def is_closeable(self):
        """Check that the canvas and contained data is safe to be closed."""
        return self.renderer.is_closeable()

    def close(self):
        """Destroy all contained data so that the window can be safely destroyed."""
        self.renderer.close()

    @property
    def world(self) -> BaseLevel:
        """The amulet-core `BaseLevel` subclass representing the "world" that is open."""
        return self._world()

    @property
    def renderer(self) -> Renderer:
        """The `Renderer` class that handles the drawable objects and draws them."""
        return self._renderer

    @property
    def dimension(self) -> Dimension:
        """The currently loaded dimension in the renderer."""
        return self.renderer.dimension

    @dimension.setter
    def dimension(self, dimension: Dimension):
        """Set the currently loaded dimension in the renderer."""
        self.renderer.dimension = dimension

    @property
    def camera(self) -> ControllableCamera:
        """A class holding the state of the camera with methods to access and modify the state."""
        return self._camera

    @property
    def buttons(self) -> ButtonInput:
        """A class that manages converting button presses into actions based on set keybinds."""
        return self._buttons

    @property
    def mouse(self) -> MouseMovement:
        """A class that manages mouse movement."""
        return self._mouse

    @property
    def selection(self) -> SelectionManager:
        """A simple class for storing the selection state."""
        return self._selection.value

    def _on_size(self, evt):
        wx.CallAfter(self._set_size)
        evt.Skip()

    def _set_size(self):
        size = self.GetClientSize() * self.GetContentScaleFactor()
        width, height = size
        self.SetCurrent(self._context)
        glViewport(0, 0, width, height)
        if height > 0:
            self.camera.aspect_ratio = width / height
        else:
            self.camera.aspect_ratio = 1



import numpy
from OpenGL.GL import (
    GL_TRIANGLES,
    glCullFace,
    GL_FRONT,
    GL_BACK,
    glDisable,
    GL_DEPTH_TEST,
    GL_LINE_STRIP,
    glEnable,
    glGetBooleanv,
    glGetIntegerv,
    GL_CULL_FACE_MODE,
)

from amulet.api.data_types import PointCoordinatesAny
from .render_selection_highlightable import RenderSelectionHighlightable
from amulet_map_editor.api.opengl.resource_pack import OpenGLResourcePack
from amulet_map_editor.api.opengl.data_types import RGBColour
from .colours import colours


class RenderSelectionEditable(RenderSelectionHighlightable):
    """A drawable selection box with additional editing controls"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        super().__init__(context_identifier, resource_pack)
        # is the locked or is it being modified. Changes the colour of the faces.
        self._locked = True

    def _init_verts(self):
        # the first 36 verts are used for the full box which is used for lines
        # the next 36 verts are used for the inset faces
        # the next 144 verts are used for the edges
        # the next 144 verts are used for the corners

        verts_per_quad = 2 * 3  # triangles * verts
        self.verts = numpy.zeros(
            (
                6 * verts_per_quad
                + 6  # original box verts (used for the lines)
                * 9
                * verts_per_quad,  # new verts
                self._vert_len,
            ),
            dtype=numpy.float32,
        )
        self.verts[:, 5:9] = self.resource_pack.texture_bounds(
            self.resource_pack.get_texture_path("amulet", "amulet_ui/selection")
        )

        self.verts[verts_per_quad * 6 :, 9:12] = self.box_tint
        self.verts[verts_per_quad * 12 : verts_per_quad * 36, 9:12] = self.edge_colour

    @property
    def highlight_colour(self) -> RGBColour:
        if self.locked:
            return colours.get("box_highlight", (0.5, 0.5, 1.0))
        else:
            return colours.get("box_highlight_move", (1.0, 0.7, 0.3))

    @property
    def edge_colour(self) -> RGBColour:
        return colours.get("box_edge", (0.5, 1.0, 1.0))

    @property
    def corner_colour(self) -> RGBColour:
        return colours.get("box_corner", (1.0, 1.0, 0.5))

    @property
    def point1_colour(self) -> RGBColour:
        return colours.get("box_point1", (0.0, 1.0, 0.0))

    @property
    def point2_colour(self) -> RGBColour:
        return colours.get("box_point2", (0.0, 0.0, 1.0))

    @property
    def locked(self) -> bool:
        """Is the selection locked or not.
        If locked (True) the highlight colour will be used, if unlocked (False) the moving colour will be used."""
        return self._locked

    @locked.setter
    def locked(self, lock: bool):
        """Set if the selection locked or not.
        If locked (True) the highlight colour will be used, if unlocked (False) the moving colour will be used."""
        self._locked = bool(lock)

    def _create_geometry_(self):
        super()._create_geometry_()

        point1, point2 = self._points - self.min + (self.min % 16)
        size = numpy.abs(point2 - point1)
        verts_per_face = 2 * 3  # triangles * verts
        # the edges of the box
        min_point, max_point = numpy.sort([point1, point2], 0).astype(numpy.float64)
        min_point -= 0.01
        max_point += 0.01
        # the edge points offset by the boundary amount.
        min_point_1 = min_point + numpy.min([numpy.ones(3), size / 4], 0)
        max_point_1 = max_point - numpy.min([numpy.ones(3), size / 4], 0)

        # down, up
        # west, east
        # north, south

        face_offset = verts_per_face * 6

        # inset faces
        for axis in ("y", "z", "x"):
            (
                self.verts[face_offset : face_offset + verts_per_face * 2, :3],
                self.verts[face_offset : face_offset + verts_per_face * 2, 3:5],
            ) = self._create_box_faces(
                (
                    min_point[0] if axis == "x" else min_point_1[0],
                    min_point[1] if axis == "y" else min_point_1[1],
                    min_point[2] if axis == "z" else min_point_1[2],
                ),
                (
                    max_point[0] if axis == "x" else max_point_1[0],
                    max_point[1] if axis == "y" else max_point_1[1],
                    max_point[2] if axis == "z" else max_point_1[2],
                ),
                up=axis == "y",
                down=axis == "y",
                north=axis == "z",
                south=axis == "z",
                west=axis == "x",
                east=axis == "x",
            )
            face_offset += verts_per_face * 2

        for y in (False, True):
            for x in (False, True):
                (
                    self.verts[face_offset : face_offset + verts_per_face * 2, :3],
                    self.verts[face_offset : face_offset + verts_per_face * 2, 3:5],
                ) = self._create_box_faces(
                    (
                        max_point_1[0] if x else min_point[0],
                        max_point_1[1] if y else min_point[1],
                        min_point_1[2],
                    ),
                    (
                        max_point[0] if x else min_point_1[0],
                        max_point[1] if y else min_point_1[1],
                        max_point_1[2],
                    ),
                    up=y,
                    down=not y,
                    west=not x,
                    east=x,
                )
                face_offset += verts_per_face * 2

        for y in (False, True):
            for z in (False, True):
                (
                    self.verts[face_offset : face_offset + verts_per_face * 2, :3],
                    self.verts[face_offset : face_offset + verts_per_face * 2, 3:5],
                ) = self._create_box_faces(
                    (
                        min_point_1[0],
                        max_point_1[1] if y else min_point[1],
                        max_point_1[2] if z else min_point[2],
                    ),
                    (
                        max_point_1[0],
                        max_point[1] if y else min_point_1[1],
                        max_point[2] if z else min_point_1[2],
                    ),
                    up=y,
                    down=not y,
                    north=not z,
                    south=z,
                )
                face_offset += verts_per_face * 2

        for x in (False, True):
            for z in (False, True):
                (
                    self.verts[face_offset : face_offset + verts_per_face * 2, :3],
                    self.verts[face_offset : face_offset + verts_per_face * 2, 3:5],
                ) = self._create_box_faces(
                    (
                        max_point_1[0] if x else min_point[0],
                        min_point_1[1],
                        max_point_1[2] if z else min_point[2],
                    ),
                    (
                        max_point[0] if x else min_point_1[0],
                        max_point_1[1],
                        max_point[2] if z else min_point_1[2],
                    ),
                    north=not z,
                    south=z,
                    west=not x,
                    east=x,
                )
                face_offset += verts_per_face * 2

        self.verts[216:360, 9:12] = self.corner_colour
        corners = point2 >= point1
        not_corners = numpy.invert(corners)
        # corners
        for y in (False, True):
            for z in (False, True):
                for x in (False, True):
                    (
                        self.verts[face_offset : face_offset + verts_per_face * 3, :3],
                        self.verts[face_offset : face_offset + verts_per_face * 3, 3:5],
                    ) = self._create_box_faces(
                        (
                            max_point_1[0] if x else min_point[0],
                            max_point_1[1] if y else min_point[1],
                            max_point_1[2] if z else min_point[2],
                        ),
                        (
                            max_point[0] if x else min_point_1[0],
                            max_point[1] if y else min_point_1[1],
                            max_point[2] if z else min_point_1[2],
                        ),
                        up=y,
                        down=not y,
                        north=not z,
                        south=z,
                        west=not x,
                        east=x,
                    )
                    if numpy.array_equal(corners, (x, y, z)):
                        self.verts[
                            face_offset : face_offset + verts_per_face * 3, 9:12
                        ] = self.point2_colour
                    elif numpy.array_equal(not_corners, (x, y, z)):
                        self.verts[
                            face_offset : face_offset + verts_per_face * 3, 9:12
                        ] = self.point1_colour
                    face_offset += verts_per_face * 3

        self.verts[:, 3:5] /= 16

        self.verts[36:72, 9:12] = self.box_tint

        indexes = numpy.zeros(6, numpy.uint8)
        if self.point2[0] > self.point1[0]:
            indexes[[4, 5]] = 0, 3
        else:
            indexes[[4, 5]] = 3, 0

        if self.point2[1] > self.point1[1]:
            indexes[[0, 1]] = 1, 4
        else:
            indexes[[0, 1]] = 4, 1

        if self.point2[2] > self.point1[2]:
            indexes[[2, 3]] = 2, 5
        else:
            indexes[[2, 3]] = 5, 2

        self.verts[36:72][
            numpy.repeat(self._highlight_edges.ravel()[indexes], 6), 9:12
        ] = self.highlight_colour

    def draw(
        self, camera_matrix: numpy.ndarray, camera_position: PointCoordinatesAny = None
    ):
        """
        Draw the selection box
        :param camera_matrix: 4x4 transformation matrix for the camera
        :param camera_position: The position of the camera. Used to flip draw direction if camera inside box.
        :return:
        """
        self._setup()
        if self._needs_rebuild:
            self._create_geometry()

        transformation_matrix = numpy.matmul(camera_matrix, self.transformation_matrix)

        depth_state = glGetBooleanv(GL_DEPTH_TEST)
        cull_state = glGetIntegerv(GL_CULL_FACE_MODE)

        # draw the lines around the boxes
        self.draw_start = 0
        self.draw_count = 36

        if depth_state:
            glDisable(GL_DEPTH_TEST)
        self._draw_mode = GL_LINE_STRIP
        super()._draw(transformation_matrix)
        if depth_state:
            glEnable(GL_DEPTH_TEST)

        if camera_position is not None:
            if camera_position in self:
                glCullFace(GL_FRONT)
            else:
                glCullFace(GL_BACK)
        self._draw_mode = GL_TRIANGLES
        self.draw_start = 36
        # 6 faces, 9 quads/face, 2 triangles/quad, 3 verts/triangle
        self.draw_count = 324
        super()._draw(transformation_matrix)

        glCullFace(cull_state)
